#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from picai_prep.data_utils import atomic_file_copy, atomic_image_write
from picai_prep.dcm2dce import Dicom2DCEConverter
from picai_prep.dcm2mha import Dicom2MHAConverter
from picai_prep.mha2nnunet import MHA2nnUNetConverter
from picai_prep.nnunet2nndet import nnunet2nndet

print("If you have questions or suggestions, feel free to open an issue " +
      "at https://github.com/DIAGNijmegen/picai_prep\n")

__all__ = [
      # Explicitly expose these functions for easier imports
      "Dicom2MHAConverter",
      "Dicom2DCEConverter",
      "MHA2nnUNetConverter",
      "nnunet2nndet",
      "atomic_image_write",
      "atomic_file_copy"
]



dcm2mha_schema = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "properties": {
        "archive": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "patient_id": {
                        "description": "Custom patient ID. If not provided, uses the 'PatientID' metadata",
                        "type": "string"
                    },
                    "study_id": {
                        "description": "Custom study ID. If not provided, uses the 'StudyInstanceUID' metadata",
                        "type": "string"
                    },
                    "path": {
                        "description": "Path to study directory. Can be relative from --input or absolute",
                        "type": "string"
                    }
                },
                "required": [
                    "patient_id",
                    "study_id",
                    "path"
                ],
                "additionalProperties": False
            }
        },
        "mappings": {
            "type": "object",
            "description": "Filter the archive using a key/value system, each study can have at most 1 mapping",
            "patternProperties": {
                "^[A-Za-z_][A-Za-z0-9_]*$": {
                    "description": "Keys can be any of the items found in metadata.json, values can be any string",
                    "type": "object"
                }
            },
            "additionalProperties": False
        },
        "options": {
            "type": "object",
            "additionalProperties": True
        }
    },
    "additionalProperties": False,
    "required": [
        "archive",
        "mappings"
    ]
}



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import os
import shutil
import subprocess
import sys
from pathlib import Path

import pytest
import SimpleITK as sitk
from numpy.testing import assert_allclose

from picai_prep.dcm2mha import (Dicom2MHACase, Dicom2MHAConverter,
                                Dicom2MHASettings, DICOMImageReader)
from picai_prep.examples.dcm2mha.sample_archive import \
    generate_dcm2mha_settings


def test_dcm2mha(
    input_dir: str = "tests/input/dcm/ProstateX",
    output_dir: str = "tests/output/mha/ProstateX",
    output_expected_dir: str = "tests/output-expected/mha/ProstateX",
):
    """
    Convert sample DICOM archive to MHA
    """
    # remove output folder (to prevent skipping the conversion)
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)

    # test usage from Python
    archive = Dicom2MHAConverter(
        input_dir=input_dir,
        output_dir=output_dir,
        dcm2mha_settings="tests/output-expected/dcm2mha_settings.json"
    )
    archive.convert()

    # compare output
    for patient_id, subject_id in [
        ("ProstateX-0000", "ProstateX-0000_07-07-2011"),
        ("ProstateX-0001", "ProstateX-0001_07-08-2011"),
    ]:
        for modality in ["t2w", "adc", "hbv", "sag", "cor"]:
            # construct paths to MHA images
            path_out = os.path.join(output_dir, patient_id, f"{subject_id}_{modality}.mha")
            path_out_expected = os.path.join(output_expected_dir, patient_id, f"{subject_id}_{modality}.mha")

            # sanity check: check if outputs exist
            assert os.path.exists(path_out), f"Could not find output file at {path_out}!"
            assert os.path.exists(path_out_expected), f"Could not find output file at {path_out_expected}!"

            # read images
            img = sitk.GetArrayFromImage(sitk.ReadImage(str(path_out)))
            img_expected = sitk.GetArrayFromImage(sitk.ReadImage(str(path_out_expected)))

            # compare images
            assert_allclose(img_expected, img)


def test_dcm2mha_commandline(
    input_dir: str = "tests/input/dcm/ProstateX",
    output_dir: str = "tests/output/mha/ProstateX",
    output_expected_dir: str = "tests/output-expected/mha/ProstateX",
):
    """
    Convert sample DICOM archive to MHA
    """
    # remove output folder (to prevent skipping the conversion)
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)

    # test usage from command line
    cmd = [
        "python", "-m", "picai_prep", "dcm2mha",
        "--input", input_dir,
        "--output", output_dir,
        "--json", "tests/output-expected/dcm2mha_settings.json",
        "--verbose", "2",
    ]

    # run command
    subprocess.check_call(cmd, shell=(sys.platform == "win32"))

    # compare output
    for patient_id, subject_id in [
        ("ProstateX-0000", "ProstateX-0000_07-07-2011"),
        ("ProstateX-0001", "ProstateX-0001_07-08-2011"),
    ]:
        for modality in ["t2w", "adc", "hbv", "sag", "cor"]:
            # construct paths to MHA images
            path_out = os.path.join(output_dir, patient_id, f"{subject_id}_{modality}.mha")
            path_out_expected = os.path.join(output_expected_dir, patient_id, f"{subject_id}_{modality}.mha")

            # sanity check: check if outputs exist
            assert os.path.exists(path_out), f"Could not find output file at {path_out}!"
            assert os.path.exists(path_out_expected), f"Could not find output file at {path_out_expected}!"

            # read images
            img = sitk.GetArrayFromImage(sitk.ReadImage(str(path_out)))
            img_expected = sitk.GetArrayFromImage(sitk.ReadImage(str(path_out_expected)))

            # compare images
            assert_allclose(img_expected, img)


def test_resolve_duplicates(
    input_dir: str = "tests/input/dcm/ProstateX",
):
    # setup case with duplicates
    case = Dicom2MHACase(
        input_dir=Path(input_dir),
        patient_id="ProstateX-0001",
        study_id="07-08-2011",
        paths=[
            "ProstateX-0001/07-08-2011/10.000000-t2tsetra-17541",
            "ProstateX-0001/07-08-2011/6.000000-t2tsetra-76610",
        ],
        settings=Dicom2MHASettings(
            mappings={
                "t2w": {
                    "SeriesDescription": [
                        "t2_tse_tra"
                    ]
                },
            }
        )
    )

    # resolve duplicates
    case.initialize()
    case.extract_metadata()
    case.apply_mappings()
    case.resolve_duplicates()

    # check if duplicates were resolved
    matched_series = [serie for serie in case.valid_series if "t2w" in serie.mappings]
    assert len(matched_series) == 1, "More than one serie after resolving duplicates!"


def test_value_match_contains(
    input_dir: str = "tests/input/dcm/ProstateX",
    patient_id="ProstateX-0001",
    study_id="07-08-2011",
):
    # setup case
    series_list = os.listdir(os.path.join(input_dir, patient_id, study_id))
    case = Dicom2MHACase(
        input_dir=Path(input_dir),
        patient_id=patient_id,
        study_id=study_id,
        paths=[
            os.path.join(patient_id, study_id, series_id) for series_id in series_list
            if os.path.isdir(os.path.join(input_dir, patient_id, study_id, series_id))
        ],
        settings=Dicom2MHASettings(
            mappings={
                "test": {
                    "SeriesDescription": [
                        ""
                    ]
                },
            },
            values_match_func="lower_strip_contains"
        )
    )

    # resolve duplicates
    case.initialize()
    case.extract_metadata()
    case.apply_mappings()

    # check if duplicates were resolved
    matched_series = [serie for serie in case.valid_series if "test" in serie.mappings]
    assert len(matched_series) == 11, "Empty lower_strip_contains should match all series!"


def test_value_match_multiple_keys(
    input_dir: str = "tests/input/dcm/ProstateX",
    patient_id="ProstateX-0001",
    study_id="07-08-2011",
):
    # setup case
    series_list = os.listdir(os.path.join(input_dir, patient_id, study_id))
    paths = [
        os.path.join(patient_id, study_id, series_id) for series_id in series_list
        if os.path.isdir(os.path.join(input_dir, patient_id, study_id, series_id))
    ]
    archive = Dicom2MHAConverter(
        input_dir=Path(input_dir),
        output_dir="",
        dcm2mha_settings={
            "archive": [
                {
                    "patient_id": patient_id,
                    "study_id": study_id,
                    "path": path
                }
                for path in paths
            ],
            "mappings": {
                "test": {
                    "SeriesDescription": [
                        ""
                    ],
                    "imagetype": [
                        "DERIVED\\PRIMARY\\DIFFUSION"
                    ]
                },
            },
            "options": {
                "values_match_func": "lower_strip_contains"
            }
        }
    )

    case = archive.cases[0]

    # resolve duplicates
    case.initialize()
    case.extract_metadata()
    case.apply_mappings()

    # check if duplicates were resolved
    matched_series = [serie for serie in case.valid_series if "test" in serie.mappings]
    assert len(matched_series) == 3, "Should find three diffusion scans!"


@pytest.mark.parametrize("input_dir", [
    "tests/input/dcm/ProstateX/ProstateX-0000/07-07-2011/4.000000-t2tsetra-00702",
    "tests/input/dcm/ProstateX/ProstateX-0000/07-07-2011/7.000000-ep2ddifftraDYNDISTADC-48780",
    "tests/input/dcm/ProstateX/ProstateX-0001/07-08-2011/11.000000-tfl3d PD reftra1.5x1.5t3-77124",
])
def test_image_reader(input_dir: str):
    """
    Verify consistency of reading SimpleITK or pydicom.
    """

    reader = DICOMImageReader(path=input_dir)
    image_sitk = reader._read_image_sitk()
    image_pydicom = reader._read_image_pydicom()

    # compare voxel values
    assert_allclose(sitk.GetArrayFromImage(image_sitk), sitk.GetArrayFromImage(image_pydicom))

    # compare physical metadata
    assert_allclose(image_sitk.GetSpacing(), image_pydicom.GetSpacing(), atol=1e-6)
    assert_allclose(image_sitk.GetOrigin(), image_pydicom.GetOrigin(), atol=1e-6)
    assert_allclose(image_sitk.GetDirection(), image_pydicom.GetDirection(), atol=1e-6)

    # compare metadata
    metadata_sitk = {key: image_sitk.GetMetaData(key) for key in image_sitk.GetMetaDataKeys()}
    metadata_pydicom = {key: image_pydicom.GetMetaData(key) for key in image_pydicom.GetMetaDataKeys()}
    keys = set(metadata_pydicom.keys()) & set(metadata_sitk.keys())
    assert len(metadata_sitk.keys()) > 1, "No metadata found with SimpleITK!"
    assert len(metadata_pydicom.keys()) > 1, "No metadata found with pydicom!"
    assert len(keys) > 1, f"No metadata found! SimpleITK: {list(metadata_sitk.keys())}, pydicom: {list(metadata_pydicom.keys())}"
    assert {k: metadata_sitk[k] for k in keys} == {k: metadata_pydicom[k] for k in keys}


@pytest.mark.xfail
@pytest.mark.parametrize("input_dir", [
    "tests/input/dcm/ProstateX/ProstateX-0001/07-08-2011/1.000000-t2localizer-75055",
])
def test_image_reader_invalid_sequence(input_dir: str):
    test_image_reader(input_dir)


@pytest.mark.parametrize("input_dir", [
    "tests/input/dcm/ProstateX-dicom-zip/ProstateX-0001/07-08-2011/8.000000-ep2ddifftraDYNDISTMIXADC-33954",
    "tests/input/dcm/ProstateX-dicom-zip/ProstateX-0001/07-08-2011/11.000000-tfl3d PD reftra1.5x1.5t3-77124",
])
def test_image_reader_dicom_zip(input_dir: str):
    reader1 = DICOMImageReader(path=input_dir)
    reader2 = DICOMImageReader(path=input_dir)

    # read image, then metadata
    image1 = reader1.image
    metadata1 = reader1.metadata

    # read metadata, then image
    metadata2 = reader2.metadata
    image2 = reader2.image

    # compare voxel values
    assert_allclose(sitk.GetArrayFromImage(image1), sitk.GetArrayFromImage(image2))

    # compare metadata
    keys = set(metadata1.keys()) & set(metadata2.keys())
    assert len(metadata1.keys()) > 1, "No metadata found with first reader!"
    assert len(metadata2.keys()) > 1, "No metadata found with second reader!"
    assert len(keys) > 1, f"No metadata found! First reader: {list(metadata1.keys())}, second reader: {list(metadata2.keys())}"
    assert {k: metadata1[k] for k in keys} == {k: metadata2[k] for k in keys}


@pytest.mark.xfail
@pytest.mark.parametrize("input_dir", [
    "tests/input/dcm/ProstateX-dicom-zip/ProstateX-0001/07-08-2011/1.000000-t2localizer-75055",
    "tests/input/dcm/ProstateX-dicom-zip/ProstateX-0001/07-08-2011/corrupt-sequence",
])
def test_image_reader_dicom_zip_invalid_sequence(input_dir: str):
    test_image_reader_dicom_zip(input_dir)


@pytest.mark.xfail
def test_image_reader_missing_slice():
    """
    Verify that a missing slice is detected.
    """
    input_dir = "tests/input/dcm/ProstateX-missing-slice/ProstateX-0000/07-07-2011/8.000000-ep2ddifftraDYNDISTCALCBVAL-83202"
    _ = DICOMImageReader(path=input_dir).image


def test_image_reader_missing_slice_okay():
    """
    Verify that an image with a missing slice can be converted.
    """
    input_dir = "tests/input/dcm/ProstateX-missing-slice/ProstateX-0000/07-07-2011/8.000000-ep2ddifftraDYNDISTCALCBVAL-83202"
    image = DICOMImageReader(path=input_dir, verify_dicom_filenames=False).image
    image = sitk.GetArrayFromImage(image)
    assert image.shape == (18, 128, 84)


def test_Dicom2MHAConverter_missing_slice_okay(
    input_dir: str = "tests/input/dcm/ProstateX-missing-slice",
    output_dir: str = "tests/output/mha/ProstateX-missing-slice",
    output_expected_dir: str = "tests/output-expected/mha/ProstateX-missing-slice",
):
    """
    Verify that an image with a missing slice can be converted with Dicom2MHAConverter
    """
    # paths
    dcm2mha_settings_path = Path(output_dir) / "dcm2mha_settings.json"

    # remove output folder (to prevent skipping the conversion)
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)

    # generate dcm2mha_settings
    dcm2mha_settings_path.parent.mkdir(parents=True, exist_ok=True)
    generate_dcm2mha_settings(
        archive_dir=input_dir,
        output_path=dcm2mha_settings_path,

    )

    # test usage from Python
    archive = Dicom2MHAConverter(
        input_dir=input_dir,
        output_dir=output_dir,
        dcm2mha_settings=dcm2mha_settings_path
    )
    archive.settings.verify_dicom_filenames = False
    archive.convert()

    # compare output
    for patient_id, subject_id in [
        ("ProstateX-0000", "ProstateX-0000_07-07-2011"),
    ]:
        for modality in ["hbv"]:
            # construct paths to MHA images
            path_out = os.path.join(output_dir, patient_id, f"{subject_id}_{modality}.mha")
            path_out_expected = os.path.join(output_expected_dir, patient_id, f"{subject_id}_{modality}.mha")

            # sanity check: check if outputs exist
            assert os.path.exists(path_out), f"Could not find output file at {path_out}!"
            assert os.path.exists(path_out_expected), f"Could not find output file at {path_out_expected}!"

            # read images
            img = sitk.GetArrayFromImage(sitk.ReadImage(str(path_out)))
            img_expected = sitk.GetArrayFromImage(sitk.ReadImage(str(path_out_expected)))

            # compare images
            assert_allclose(img_expected, img)



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import os
import shutil
from pathlib import Path
from typing import Union

import SimpleITK as sitk

PathLike = Union[str, Path]


def atomic_image_write(
    image: sitk.Image,
    path: PathLike,
    backup_existing_file: bool = False,
    compress: bool = True,
    mkdir: bool = False
):
    """
    Safely write image to disk, by:
    1. Writing the image to a temporary file: path/to/tmp_[filename]
    2. IF writing succeeded:
    2a. (optional) rename existing file to path/to/backup_[filename]
    2b. rename file to target name, which is an atomic operation
    This way, no partially written files can exist at the target path (except for temporary files)
    """
    path = Path(path)

    if mkdir:
        os.makedirs(path.parent, exist_ok=True)

    # save image to temporary file
    path_tmp = path.with_name(f"tmp_{path.name}")
    sitk.WriteImage(image, path_tmp.as_posix(), useCompression=compress)

    # backup existing file?
    if backup_existing_file and path.exists():
        dst_path_bak = path.with_name(f"backup_{path.name}")
        if dst_path_bak.exists():
            raise FileExistsError(f"Existing backup file found at {dst_path_bak}.")
        os.replace(path, dst_path_bak)

    # rename temporary file
    os.replace(path_tmp, path)


def atomic_file_copy(
    src_path: PathLike,
    dst_path: PathLike,
    backup_existing_file: bool = False,
    mkdir: bool = False
):
    """
    Safely copy file, by:
    1. Writing the file to a temporary file: path/to/.tmp.[filename]
    2. IF writing succeeded:
    2a. (optional) backup existing file to path/to/backup.[filename]
    2b. rename file to target name, which is an atomic operation
    This way, no partially written files can exist at the target path (except for temporary files)
    """
    dst_path = Path(dst_path)

    if mkdir:
        os.makedirs(dst_path.parent, exist_ok=True)

    # copy file to temporary location
    dst_path_tmp = dst_path.with_name(f".tmp.{dst_path.name}")
    try:
        shutil.copy(src_path, dst_path_tmp)
    except PermissionError:
        shutil.copyfile(src_path, dst_path_tmp)

    # backup existing file?
    if backup_existing_file and dst_path.exists():
        dst_path_bak = dst_path.with_name(f"backup.{dst_path.name}")
        if dst_path_bak.exists():
            raise FileExistsError(f"Existing backup file found at {dst_path_bak}.")
        os.replace(dst_path, dst_path_bak)

    # rename temporary file
    os.replace(dst_path_tmp, dst_path)



dicom_tags = {
    "accessionnumber": "0008|0050",
    "acquisitiondate": "0008|0022",
    "acquisitionmatrix": "0018|1310",
    "acquisitiontime": "0008|0032",
    "additionalpatienthistory": "0010|21b0",
    "bitsallocated": "0028|0100",
    "bitsstored": "0028|0101",
    "bodypartexamined": "0018|0015",
    "columns": "0028|0011",
    "contentdate": "0008|0023",
    "contenttime": "0008|0033",
    "deidentificationmethod": "0012|0063",
    "deviceserialnumber": "0018|1000",
    "diffusionbvalue": "0018|9087",
    "echotime": "0018|0081",
    "echotrainlength": "0018|0091",
    "flipangle": "0018|1314",
    "frameofreferenceuid": "0020|0052",
    "imagednucleus": "0018|0085",
    "imageorientationpatient": "0020|0037",
    "imagepositionpatient": "0020|0032",
    "imagetype": "0008|0008",
    "imagingfrequency": "0018|0084",
    "inplanephaseencodingdirection": "0018|1312",
    "instancecreationdate": "0008|0012",
    "instancecreationtime": "0008|0013",
    "institutionaldepartmentname": "0008|1040",
    "largestimagepixelvalue": "0028|0107",
    "magneticfieldstrength": "0018|0087",
    "manufacturer": "0008|0070",
    "manufacturersmodelname": "0008|1090",
    "modality": "0008|0060",
    "mracquisitiontype": "0018|0023",
    "numberofaverages": "0018|0083",
    "numberofphaseencodingsteps": "0018|0089",
    "patientid": "0010|0020",
    "patientidentityremoved": "0012|0062",
    "patientposition": "0018|5100",
    "patientsage": "0010|1010",
    "patientsbirthdate": "0010|0030",
    "patientssex": "0010|0040",
    "percentphasefieldofview": "0018|0094",
    "percentsampling": "0018|0093",
    "performedprocedurestepdescription": "0040|0254",
    "performedprocedurestepstartdate": "0040|0244",
    "pixelspacing": "0028|0030",
    "protocolname": "0018|1030",
    "repetitiontime": "0018|0080",
    "rescaleintercept": "0028|1052",
    "rescaleslope": "0028|1053",
    "rows": "0028|0010",
    "samplesperpixel": "0028|0002",
    "scanningsequence": "0018|0020",
    "scanoptions": "0018|0022",
    "sequencename": "0018|0024",
    "sequencevariant": "0018|0021",
    "seriesdate": "0008|0021",
    "seriesdescription": "0008|103e",
    "seriesinstanceuid": "0020|000e",
    "seriestime": "0008|0031",
    "slicelocation": "0020|1041",
    "slicethickness": "0018|0050",
    "smallestimagepixelvalue": "0028|0106",
    "sopclassuid": "0008|0016",
    "sopinstanceuid": "0008|0018",
    "spacingbetweenslices": "0018|0088",
    "specificcharacterset": "0008|0005",
    "stationname": "0008|1010",
    "studydate": "0008|0020",
    "studydescription": "0008|1030",
    "studyid": "0020|0010",
    "studyinstanceuid": "0020|000d",
    "studytime": "0008|0030",
    "variableflipangleflag": "0018|1315",
    "windowcenter": "0028|1050",
    "windowwidth": "0028|1051"
}



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import json
import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import List, Optional

import pytest
import SimpleITK as sitk
from numpy.testing import assert_allclose

from picai_prep.data_utils import PathLike
from picai_prep.examples.mha2nnunet import picai_archive
from picai_prep.mha2nnunet import MHA2nnUNetConverter
from picai_prep.preprocessing import resample_to_reference_scan


def test_mha2nnunet(
    input_dir: PathLike = "tests/output-expected/mha/ProstateX",
    annotations_dir: PathLike = "tests/input/annotations/ProstateX",
    output_dir: PathLike = "tests/output/nnUNet_raw_data",
    output_expected_dir: PathLike = "tests/output-expected/nnUNet_raw_data",
    settings_path: PathLike = "tests/output-expected/mha2nnunet_settings.json",
    task_name: str = "Task100_test",
    subject_list: Optional[List[str]] = None,
):
    """
    Convert sample MHA archive to nnUNet raw data format
    """
    if subject_list is None:
        subject_list = [
            "ProstateX-0000_07-07-2011",
            "ProstateX-0001_07-08-2011",
        ]

    # convert input paths to Path
    input_dir = Path(input_dir)
    annotations_dir = Path(annotations_dir)
    output_dir = Path(output_dir)
    output_expected_dir = Path(output_expected_dir)
    task_dir = output_dir / task_name

    # remove output folder (to prevent skipping the conversion)
    if os.path.exists(task_dir):
        shutil.rmtree(task_dir)

    # convert MHA archive to nnUNet raw data
    archive = MHA2nnUNetConverter(
        output_dir=output_dir.as_posix(),
        scans_dir=input_dir.as_posix(),
        annotations_dir=annotations_dir.as_posix(),
        mha2nnunet_settings=settings_path
    )
    archive.convert()
    archive.create_dataset_json()

    # check dataset.json
    path_out = task_dir / "dataset.json"
    path_out_expected = output_expected_dir / task_name / "dataset.json"
    with open(path_out) as fp1, open(path_out_expected) as fp2:
        assert json.load(fp1) == json.load(fp2)

    # compare output
    for subject_id in subject_list:
        print(f"Checking case {subject_id}...")
        case_origin, case_direction = None, None

        for modality in ["0000", "0001", "0002"]:
            # construct paths to MHA images
            path_out = task_dir / "imagesTr" / f"{subject_id}_{modality}.nii.gz"
            path_out_expected = output_expected_dir / task_name / "imagesTr" / f"{subject_id}_{modality}.nii.gz"

            # sanity check: check if outputs exist
            assert path_out.exists(), f"Could not find output file at {path_out}!"
            assert path_out_expected.exists(), f"Could not find output file at {path_out_expected}!"

            # read images
            img = sitk.ReadImage(str(path_out))
            img_expected = sitk.ReadImage(str(path_out_expected))

            # compare images
            assert_allclose(sitk.GetArrayFromImage(img_expected), sitk.GetArrayFromImage(img))

            # check origin and direction (nnUNet checks this too)
            if case_origin is None:
                case_origin = img.GetOrigin()
            else:
                assert case_origin == img.GetOrigin(), "Origin must match between sequences!"
            if case_direction is None:
                case_direction = img.GetDirection()
            else:
                assert case_direction == img.GetDirection(), "Direction must match between sequences!"

        # check annotation
        # construct paths
        path_out = task_dir / "labelsTr" / f"{subject_id}.nii.gz"
        path_out_expected = output_expected_dir / task_name / "labelsTr" / f"{subject_id}.nii.gz"

        # sanity check: check if outputs exist
        assert path_out.exists(), f"Could not find output file at {path_out}!"
        assert path_out_expected.exists(), f"Could not find output file at {path_out_expected}!"

        # read images
        img = sitk.GetArrayFromImage(sitk.ReadImage(str(path_out)))
        img_expected = sitk.GetArrayFromImage(sitk.ReadImage(str(path_out_expected)))

        # compare images
        assert_allclose(img_expected, img)


def test_mha2nnunet_commandline(
    input_dir: PathLike = "tests/output-expected/mha/ProstateX",
    annotations_dir: PathLike = "tests/input/annotations/ProstateX",
    output_dir: PathLike = "tests/output/nnUNet_raw_data",
    output_expected_dir: PathLike = "tests/output-expected/nnUNet_raw_data",
    settings_path: PathLike = "tests/output-expected/mha2nnunet_settings.json",
    task_name: str = "Task100_test",
    subject_list: Optional[List[str]] = None,
):
    """
    Convert sample MHA archive to nnUNet raw data format
    """
    if subject_list is None:
        subject_list = [
            "ProstateX-0000_07-07-2011",
            "ProstateX-0001_07-08-2011",
        ]

    # convert input paths to Path
    input_dir = Path(input_dir)
    annotations_dir = Path(annotations_dir)
    output_dir = Path(output_dir)
    output_expected_dir = Path(output_expected_dir)
    task_dir = output_dir / task_name

    # remove output folder (to prevent skipping the conversion)
    if os.path.exists(task_dir):
        shutil.rmtree(task_dir)

    # test usage from command line
    cmd = [
        "python", "-m", "picai_prep", "mha2nnunet",
        "--input", input_dir.as_posix(),
        "--output", output_dir.as_posix(),
        "--annotations", annotations_dir.as_posix(),
        "--json", settings_path,
        "--verbose", "2",
    ]

    # run command
    subprocess.check_call(cmd, shell=(sys.platform == 'win32'))

    # check dataset.json
    path_out = task_dir / "dataset.json"
    path_out_expected = output_expected_dir / task_name / "dataset.json"
    with open(path_out) as fp1, open(path_out_expected) as fp2:
        assert json.load(fp1) == json.load(fp2)

    # compare output
    for subject_id in subject_list:
        print(f"Checking case {subject_id}...")
        case_origin, case_direction = None, None

        for modality in ["0000", "0001", "0002"]:
            # construct paths to MHA images
            path_out = task_dir / "imagesTr" / f"{subject_id}_{modality}.nii.gz"
            path_out_expected = output_expected_dir / task_name / "imagesTr" / f"{subject_id}_{modality}.nii.gz"

            # sanity check: check if outputs exist
            assert path_out.exists(), f"Could not find output file at {path_out}!"
            assert path_out_expected.exists(), f"Could not find output file at {path_out_expected}!"

            # read images
            img = sitk.ReadImage(str(path_out))
            img_expected = sitk.ReadImage(str(path_out_expected))

            # compare images
            assert_allclose(sitk.GetArrayFromImage(img_expected), sitk.GetArrayFromImage(img))

            # check origin and direction (nnUNet checks this too)
            if case_origin is None:
                case_origin = img.GetOrigin()
            else:
                assert case_origin == img.GetOrigin(), "Origin must match between sequences!"
            if case_direction is None:
                case_direction = img.GetDirection()
            else:
                assert case_direction == img.GetDirection(), "Direction must match between sequences!"

        # check annotation
        # construct paths
        path_out = task_dir / "labelsTr" / f"{subject_id}.nii.gz"
        path_out_expected = output_expected_dir / task_name / "labelsTr" / f"{subject_id}.nii.gz"

        # sanity check: check if outputs exist
        assert path_out.exists(), f"Could not find output file at {path_out}!"
        assert path_out_expected.exists(), f"Could not find output file at {path_out_expected}!"

        # read images
        img = sitk.GetArrayFromImage(sitk.ReadImage(str(path_out)))
        img_expected = sitk.GetArrayFromImage(sitk.ReadImage(str(path_out_expected)))

        # compare images
        assert_allclose(img_expected, img)


def test_mha2nnunet_inference(
    input_dir: PathLike = "tests/output-expected/mha/ProstateX",
    output_dir: PathLike = "tests/output/nnUNet_raw_data",
    output_expected_dir: PathLike = "tests/output-expected/nnUNet_raw_data",
    settings_path: PathLike = "tests/output-expected/mha2nnunet_inference_settings.json",
    task_name: str = "Task100_test",
    subject_list: Optional[List[str]] = None,
):
    """
    Convert sample MHA archive to nnUNet raw data format (images only)
    """
    if subject_list is None:
        subject_list = [
            "ProstateX-0000_07-07-2011",
            "ProstateX-0001_07-08-2011",
        ]

    # convert input paths to Path
    input_dir = Path(input_dir)
    output_dir = Path(output_dir)
    output_expected_dir = Path(output_expected_dir)
    task_dir = output_dir / task_name

    # remove output folder (to prevent skipping the conversion)
    if os.path.exists(task_dir):
        shutil.rmtree(task_dir)

    # convert MHA archive to nnUNet raw data
    archive = MHA2nnUNetConverter(
        output_dir=output_dir.as_posix(),
        scans_dir=input_dir.as_posix(),
        scans_out_dirname='imagesTs',
        mha2nnunet_settings=settings_path
    )
    archive.convert()

    # for inference, no dataset.json should be created
    path_out = output_dir / task_name / "dataset.json"
    assert not path_out.exists(), "dataset.json should not be generated in inference mode!"

    # compare output
    for subject_id in subject_list:
        for modality in ["0000", "0001", "0002"]:
            # construct paths to MHA images
            path_out = output_dir / task_name / "imagesTs" / f"{subject_id}_{modality}.nii.gz"
            path_out_expected = output_expected_dir / task_name / "imagesTr" / f"{subject_id}_{modality}.nii.gz"

            # sanity check: check if outputs exist
            assert path_out.exists(), f"Could not find output file at {path_out}!"
            assert path_out_expected.exists(), f"Could not find output file at {path_out_expected}!"

            # read images
            img = sitk.GetArrayFromImage(sitk.ReadImage(str(path_out)))
            img_expected = sitk.GetArrayFromImage(sitk.ReadImage(str(path_out_expected)))

            # compare images
            assert_allclose(img_expected, img)

    # assert no annotations were converted
    path_out = output_dir / task_name / "labelsTs" / f"{subject_id}_{modality}.nii.gz"
    assert not os.path.exists(path_out), "Annotation should not be converted for inference mha2nnunet"

    path_out = output_dir / task_name / "labelsTr" / f"{subject_id}_{modality}.nii.gz"
    assert not os.path.exists(path_out), "Annotation should not be converted for inference mha2nnunet"


def test_mha2nnunet_picai(
    input_dir: PathLike = "tests/input/mha/picai",
    annotations_dir: PathLike = "tests/input/annotations/picai/csPCa_lesion_delineations/human_expert/resampled",
    output_dir: PathLike = "tests/output/nnUNet_raw_data",
    output_expected_dir: PathLike = "tests/output-expected/nnUNet_raw_data",
    settings_path: PathLike = "tests/output-expected/mha2nnunet_inference_settings_picai.json",
    task_name: str = "Task2201_picai_baseline",
):
    """
    Test mha2nnunet preprocessing with tricky cases from the PI-CAI: Public Training and Development Dataset.
    These cases are typicaly very off-centre, and contain different field-of-views between sequences.

    Notes:
    - 10032_1000032 has a different field-of-view between axial T2-weighted and diffusion scans.
    - 10059_1000059 has a padded T2-weighted scan, and different field-of-view between sequences.
    - 10699_1000715_0002.nii.gz contains all diffusion scans, which is an accident of the released MHA scan.
    - 10730_1000746 is similar to 10059_1000059, but padded on the other side.
    - 10868_1000884 is pretty normal. T2-weighted has 21 slices, while diffusion has 19, which is representative.
    - except for 10699_1000715_0002, all cases in output-expected are aligned between sequences.
    """
    # remove output folder (to prevent skipping the conversion)
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)

    picai_archive.generate_mha2nnunet_settings(
        archive_dir=input_dir,
        annotations_dir=annotations_dir,
        output_path=settings_path
    )

    test_mha2nnunet(
        input_dir=input_dir,
        annotations_dir=annotations_dir,
        output_dir=output_dir,
        output_expected_dir=output_expected_dir,
        settings_path=settings_path,
        task_name=task_name,
        subject_list=[
            "10032_1000032",
            "10059_1000059",
            "10699_1000715",
            "10730_1000746",
            "10868_1000884",
        ]
    )


@pytest.mark.parametrize("subject_id", [
    "ProstateX-0000_07-07-2011",
    "ProstateX-0001_07-08-2011",
])
def test_resample_to_reference_scan(
    subject_id: str,
    input_dir: PathLike = "tests/output-expected/mha/ProstateX",
    annotations_dir: PathLike = "tests/input/annotations/ProstateX",
    output_expected_dir: PathLike = "tests/output-expected/nnUNet_raw_data",
    task_name: str = "Task100_test",
):
    """
    Test resample_to_reference_scan by comparing the annotation with
    annotation -> nnUNet preprocessed -> resample_to_reference_scan(annotation)
    """
    # read images
    lbl = sitk.ReadImage(os.path.join(annotations_dir, f"{subject_id}.nii.gz"))
    lbl_prep = sitk.ReadImage(os.path.join(output_expected_dir, task_name, "labelsTr", f"{subject_id}.nii.gz"))
    reference_scan_original = sitk.ReadImage(os.path.join(input_dir, subject_id.split("_")[0], f"{subject_id}_t2w.mha"))
    reference_scan_preprocessed = sitk.ReadImage(os.path.join(output_expected_dir, task_name, "imagesTr", f"{subject_id}_0000.nii.gz"))

    # translate preprocessed label back to original scan's physical space
    lbl_back = resample_to_reference_scan(
        image=lbl_prep,
        reference_scan_original=reference_scan_original,
        reference_scan_preprocessed=reference_scan_preprocessed,
        interpolator=sitk.sitkNearestNeighbor,
    )

    # compare lbl->prep->back with original lbl
    assert_allclose(sitk.GetArrayFromImage(lbl_back), sitk.GetArrayFromImage(lbl))


if __name__ == "__main__":
    test_resample_to_reference_scan("ProstateX-0001_07-08-2011")



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import json
import os
from pathlib import Path
from typing import Dict

from picai_prep.data_utils import PathLike
from tqdm import tqdm


def generate_dcm2mha_settings(
    archive_dir: PathLike,
    output_path: PathLike,
    mappings: Dict = None,
    **kwargs
):
    """
    Create dcm2mha_settings.json for a DICOM archive assuming the following structure:
    /path/to/archive/
    ├── [patient UID]/
        ├── [study UID]/
            ├── [series UID]/
                ├── slice-1.dcm
                ...
                ├── slice-n.dcm

    Parameters
    ----------
    archive_dir:
        path to DICOM archive
    output_path:
        path to store DICOM->MHA settings JSON to (parent folder should exist)
    mappings:
        mapping defining which series within a case is converted, based on metadata tags

    Other Parameters
    ----------------
    num_threads: int, default: 4
        number of threads to use for multiprocessing
    verify_dicom_filenames: bool, default: True
        explicitly verify dicom filenames as a sanity check
    allow_duplicates: bool, default: False
        when multiple series apply to a mapping, convert all
    """

    archive_list = []
    archive_dir = Path(archive_dir)

    # traverse DICOM archive
    for patient_id in tqdm(sorted(os.listdir(archive_dir))):
        # traverse each patient's studies
        patient_dir: Path = archive_dir / patient_id
        if not patient_dir.is_dir():
            continue

        # collect list of available studies
        for study_id in sorted(os.listdir(patient_dir)):
            # traverse each study's sequences
            study_dir = patient_dir / study_id
            if not study_dir.is_dir():
                continue

            for series_id in sorted(os.listdir(study_dir)):
                # construct path to series folder
                path = Path(patient_id, study_id, series_id)

                if not (study_dir / series_id).is_dir():
                    continue

                # store info
                archive_list += [{
                    "patient_id": patient_id,
                    "study_id": study_id,
                    "path": path.as_posix(),
                }]

    if not mappings:
        mappings = {
            "t2w": {
                "SeriesDescription": [
                    "t2_tse_tra",
                ]
            },
            "cor": {
                "SeriesDescription": [
                    "t2_tse_cor"
                ]
            },
            "sag": {
                "SeriesDescription": [
                    "t2_tse_sag"
                ]
            },
            "adc": {
                "SeriesDescription": [
                    "ep2d_diff_tra_DYNDIST_MIX_ADC",
                    "ep2d_diff_tra_DYNDIST_ADC",
                ]
            },
            "hbv": {
                "SeriesDescription": [
                    "ep2d_diff_tra_DYNDIST_MIXCALC_BVAL",
                    "ep2d_diff_tra_DYNDISTCALC_BVAL",
                ]
            }
        }

    archive = {
        "mappings": mappings,
        "archive": archive_list
    }
    if kwargs:
        archive["options"] = kwargs

    if not len(archive_list):
        raise ValueError("Did not find any DICOM series, aborting.")

    with open(output_path, "w") as fp:
        json.dump(archive, fp, indent=4)

    print(f""""
    Saved dcm2mha_settings to {output_path}, with {len(archive_list)} DICOM series.
    """)


if __name__ == '__main__':
    generate_dcm2mha_settings()



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import json
import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import jsonschema
import SimpleITK as sitk

from picai_prep.converter import Case, Converter
from picai_prep.data_utils import PathLike, atomic_image_write
from picai_prep.preprocessing import PreprocessingSettings, Sample
from picai_prep.utilities import mha2nnunet_schema, plural


@dataclass
class MHA2nnUNetSettings:
    dataset_json: Dict[str, Any]
    preprocessing: PreprocessingSettings
    scans_dirname: PathLike = "imagesTr"
    annotation_dirname: PathLike = "labelsTr"
    annotation_preprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    annotation_postprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    segmentation_preprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    segmentation_postprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    scan_preprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    scan_postprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    case_preprocess_func: Optional[Callable[["Sample"], "Sample"]] = None
    case_postprocess_func: Optional[Callable[["Sample"], "Sample"]] = None
    num_threads: int = 4
    verbose: int = 1

    @property
    def task_name(self) -> str:
        return self.dataset_json['task']


@dataclass
class _MHA2nnUNetCaseBase:
    scans_dir: Path
    annotations_dir: Path
    segmentations_dir: Path
    scan_paths: List[Path]
    settings: MHA2nnUNetSettings


@dataclass
class MHA2nnUNetCase(Case, _MHA2nnUNetCaseBase):
    annotation_path: Optional[Path] = None
    segmentation_path: Optional[Path] = None
    skip_conversion: bool = False
    verified_scan_paths: List[Path] = field(default_factory=list)

    def __post_init__(self) -> None:
        if self.annotations_dir and self.annotation_path:
            self.annotation_path = self.annotations_dir / self.annotation_path
        if self.segmentations_dir and self.segmentation_path:
            self.segmentation_path = self.segmentations_dir / self.segmentation_path

    def convert_item(self, scans_out_dir: Path, annotations_out_dir: Path) -> None:
        self.initialize(scans_out_dir, annotations_out_dir)
        if self.skip_conversion:
            return
        self.process_and_write(scans_out_dir, annotations_out_dir)

    def initialize(self, scans_out_dir: Path, annotations_out_dir: Path):
        # check if all output paths exist
        if self.output_files_exist(scans_out_dir=scans_out_dir, annotations_out_dir=annotations_out_dir):
            self.skip_conversion = True
            return

        self.write_log(f'Importing {plural(len(self.scan_paths), "scan")}')

        missing_paths = []
        for scan_path in self.scan_paths:
            # check (relative) path of input scans
            path = self.scans_dir / scan_path
            if not path.exists():
                missing_paths.append(path)
                continue

            self.verified_scan_paths.append(path)
            self.write_log(f'\t+ ({len(self.verified_scan_paths)}) {path}')

        if len(missing_paths) > 0:
            raise FileNotFoundError(','.join([str(p) for p in missing_paths]))

        if self.annotation_path:
            self.write_log('Importing annotation')
            if not self.annotation_path.exists():
                raise FileNotFoundError(self.annotation_path)

            self.write_log(f'\t+ {self.annotation_path}')
        if self.segmentation_path:
            self.write_log('Importing segmentation')
            if not self.segmentation_path.exists():
                raise FileNotFoundError(self.segmentation_path)

            self.write_log(f'\t+ {self.segmentation_path}')

    def output_files_exist(self, scans_out_dir: Path, annotations_out_dir: Path) -> bool:
        """
        Check whether all preprocessed scans (and annotation) already exist.
        """
        destination_paths = [
            scans_out_dir / f"{self.subject_id}_{i:04d}.nii.gz"
            for i in range(len(self.scan_paths))
        ]
        if self.annotation_path:
            destination_paths.append(annotations_out_dir / f"{self.subject_id}.nii.gz")
        if all(path.exists() for path in destination_paths):
            return True

    def process_and_write(self, scans_out_dir: Path, annotations_out_dir: Path) -> None:
        """
        Read scans (and annotation), perform preprocessing and write to disk.
        """
        self.write_log(
            f'Writing {plural(len(self.verified_scan_paths), "scan")}'
            + ' including annotation' if self.annotation_path else ''
        )

        scans = [sitk.ReadImage(path.as_posix()) for path in self.verified_scan_paths]
        lbl = sitk.ReadImage(self.annotation_path.as_posix()) if self.annotation_path else None
        seg = sitk.ReadImage(self.segmentation_path.as_posix()) if self.segmentation_path else None

        # set up Sample
        sample = Sample(
            scans=scans,
            lbl=lbl,
            seg=seg,
            settings=self.settings.preprocessing,
            lbl_preprocess_func=self.settings.annotation_preprocess_func,
            lbl_postprocess_func=self.settings.annotation_postprocess_func,
            seg_preprocess_func=self.settings.segmentation_preprocess_func,
            seg_postprocess_func=self.settings.segmentation_postprocess_func,
            scan_preprocess_func=self.settings.scan_preprocess_func,
            scan_postprocess_func=self.settings.scan_postprocess_func,
            case_preprocess_func=self.settings.case_preprocess_func,
            case_postprocess_func=self.settings.case_postprocess_func,
            name=self.subject_id
        )

        # perform preprocessing
        sample.preprocess()

        # write images
        for i, scan in enumerate(sample.scans):
            destination_path = scans_out_dir / f"{self.subject_id}_{i:04d}.nii.gz"
            atomic_image_write(scan, path=destination_path, mkdir=True)
            self.write_log(f'Wrote image to {destination_path}')

        if lbl:
            destination_path = annotations_out_dir / f"{self.subject_id}.nii.gz"
            atomic_image_write(sample.lbl, path=annotations_out_dir / f"{self.subject_id}.nii.gz", mkdir=True)
            self.write_log(f'Wrote annotation to {destination_path}')

    def compile_log(self) -> Optional[str]:
        if self.settings.verbose == 0:
            # logging is disabled
            return None

        if self.skip_conversion and self.settings.verbose >= 2:
            return f"Skipping {self.subject_id}, already converted."

        if self.is_valid:
            if self.settings.verbose >= 2:
                # conversion was successful and verbose logging is enabled
                return '\n'.join(['=' * 120,
                                  f'CASE {self.subject_id}',
                                  f'\tPATIENT ID\t{self.patient_id}',
                                  f'\tSTUDY ID\t{self.study_id}\n',
                                  *self._log])
            else:
                # conversion was successful and short logging is enabled
                return '\n'.join([f'CASE {self.subject_id} successfully converted'])
        else:
            # conversion failed, log everything
            return '\n'.join(['=' * 120,
                              f'CASE {self.subject_id}',
                              f'\tPATIENT ID\t{self.patient_id}',
                              f'\tSTUDY ID\t{self.study_id}\n',
                              *self._log,
                              'Error:',
                              self.error_trace])


class MHA2nnUNetConverter(Converter):
    def __init__(
        self,
        output_dir: PathLike,
        scans_dir: PathLike,
        mha2nnunet_settings: Union[PathLike, Dict],
        scans_out_dirname: str = 'imagesTr',
        annotations_dir: Optional[PathLike] = None,
        segmentations_dir: Optional[PathLike] = None,
        annotations_out_dirname: Optional[str] = 'labelsTr',
    ):
        """
        Convert an MHA Archive to an nnUNet Raw Data Archive.
        See https://github.com/DIAGNijmegen/picai_prep for additional documentation.

        Parameters
        ----------
        output_dir: PathLike
            path to store the resulting nnUNet archive.
        scans_dir: PathLike
            directory name to store scans in, relative to `output_dir`.
        scans_out_dirname: str, default: 'imagesTr'
            dirname to store scan output, will be a direct descendant of `output_dir`.
        annotations_dir: PathLike
            path to the annotations archive. Used as base path for the relative paths of the archive items.
        annotations_out_dirname: str, default: 'labelsTr'
            dirname to store annotation output, will be a direct descendant of `output_dir`.
        mha2nnunet_settings: Union[PathLike, Dict]
            object with cases, nnUNet-shaped dataset.json and optional parameters.
            May be a dictionary containing mappings, dataset.json, and optionally options,
            or a path to a JSON file with these elements.
            - dataset_json: see nnU-Net's dataset conversion on details for the dataset.json file:
                https://github.com/MIC-DKFZ/nnUNet/blob/master/documentation/dataset_conversion.md
            - cases: list of objects. Each case should be an object with a patient_id,
                study_id, relative paths to scans and optionally a path to an annotation
            - options: (optional)
                - num_threads: number of multithreading threads.
                    Default: 4.
                - verbose: control logfile verbosity. 0 does not output a logfile,
                    1 logs cases which have critically failed, 2 logs all cases (may lead to
                    very large log files)
                    Default: 1
        """
        if isinstance(mha2nnunet_settings, (Path, str)):
            with open(mha2nnunet_settings) as fp:
                mha2nnunet_settings = json.load(fp)

        # validate and parse settings
        jsonschema.validate(mha2nnunet_settings, mha2nnunet_schema, cls=jsonschema.Draft7Validator)
        self.settings = MHA2nnUNetSettings(
            dataset_json=mha2nnunet_settings['dataset_json'],
            preprocessing=PreprocessingSettings(**mha2nnunet_settings['preprocessing']),
            **mha2nnunet_settings.get('options', {})
        )

        # set up paths and create output directory
        self.scans_dir = Path(scans_dir)
        self.annotations_dir = Path(annotations_dir) if annotations_dir else None
        self.segmentations_dir = Path(segmentations_dir) if segmentations_dir else None

        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

        self.scans_out_dir = output_dir / self.settings.task_name / scans_out_dirname
        self.annotations_out_dir = output_dir / self.settings.task_name / annotations_out_dirname if annotations_dir else None

        # initialize cases to convert
        self.cases = self._init_cases(mha2nnunet_settings['archive'])

        # set up logfile
        self.initialize_log(output_dir, self.settings.verbose)

    def _init_cases(self, archive: List[Dict[str, Any]]) -> List[MHA2nnUNetCase]:
        return [
            MHA2nnUNetCase(
                scans_dir=self.scans_dir,
                annotations_dir=self.annotations_dir,
                segmentations_dir=self.segmentations_dir,
                settings=self.settings,
                **kwargs,
            )
            for kwargs in archive
        ]

    def convert(self):
        self._convert(
            title='MHA2nnUNet',
            cases=self.cases,
            parameters={
                'scans_out_dir': self.scans_out_dir,
                'annotations_out_dir': self.annotations_out_dir
            },
            num_threads=self.settings.num_threads,
        )

    def _prepare_dataset_paths(self):
        """Prepare paths to scans and annotation in nnU-Net dataset.json format"""
        if self.annotations_out_dir is None:
            return [
                {
                    "image": f"./{self.scans_out_dir.name}/{case.subject_id}.nii.gz"
                }
                for case in self.valid_cases
            ]
        else:
            return [
                {
                    "image": f"./{self.scans_out_dir.name}/{case.subject_id}.nii.gz",
                    "label": f"./{self.annotations_out_dir.name}/{case.subject_id}.nii.gz"
                }
                for case in self.valid_cases
            ]

    def create_dataset_json(self, path: PathLike = 'dataset.json', is_testset: bool = False) -> Dict:
        """
        Create dataset.json for nnUNet raw data archive.

        Parameters
        ----------
        path: PathLike, default: nnU-Net raw data archive folder / task / dataset.json
            path to save dataset info to. If None, will not output a file.
        is_testset: bool, default: False
            whether this conversation was a test set or a training set.

        Returns
        -------
        dataset : dict
            contents of dataset.json
        """
        if path is None:
            return

        dataset_path = self.scans_out_dir.parent / path
        if dataset_path.exists():
            logging.info(f"Dataset info already exists at {dataset_path}, saving to ...-conflict.json")
            dataset_path = dataset_path.with_stem(dataset_path.stem + "-conflict")

        logging.info(f'Saving dataset info to {dataset_path}')

        # use contents of archive->dataset_json as starting point
        dataset_settings = self.settings.dataset_json
        if 'name' not in dataset_settings:
            dataset_settings['name'] = '_'.join(self.settings.task_name.split('_')[1:])

        if is_testset:
            dataset_settings["numTest"] = len(self.cases)
            dataset_settings["test"] = self._prepare_dataset_paths()
            if "numTraining" not in dataset_settings:
                dataset_settings["numTraining"] = 0
                dataset_settings["training"] = []
        else:
            dataset_settings['numTraining'] = len(self.cases)
            dataset_settings["training"] = self._prepare_dataset_paths()
            if "numTest" not in dataset_settings:
                dataset_settings["numTest"] = 0
                dataset_settings["test"] = []

        with open(dataset_path, 'w') as fp:
            json.dump(dataset_settings, fp, indent=4)

        return dataset_settings

    @property
    def valid_cases(self):
        return [case for case in self.cases if case.is_valid]



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json
import logging
import os
import re
import tempfile
import traceback
import zipfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union

import jsonschema
import numpy as np
import pydicom
import pydicom.errors
import SimpleITK as sitk

from picai_prep.converter import Case, Converter
from picai_prep.data_utils import PathLike, atomic_image_write
from picai_prep.errors import (ArchiveItemPathNotFoundError,
                               CriticalErrorInSiblingError,
                               MissingDICOMFilesError, NoMappingsApplyError,
                               UnreadableDICOMError)
from picai_prep.utilities import dcm2mha_schema, dicom_tags, plural

Metadata = Dict[str, str]
Mapping = Dict[str, List[str]]
Mappings = Dict[str, Mapping]


@dataclass
class Dicom2MHASettings:
    mappings: Dict[str, Dict[str, List[str]]]
    verify_dicom_filenames: bool = True
    allow_duplicates: bool = False
    metadata_match_func: Optional[Callable[[Metadata, Mappings], bool]] = None
    values_match_func: Union[str, Callable[[str, str], bool]] = "lower_strip_equals"
    scan_postprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    num_threads: int = 4
    verbose: int = 1

    def __post_init__(self):
        # Validate the mappings
        for mapping_name, mapping in self.mappings.items():
            for key, values in mapping.items():
                if not isinstance(values, list):
                    raise ValueError(f'Mapping {mapping_name} has non-list values for key {key}: {values}')
                if not all([isinstance(value, str) for value in values]):
                    raise ValueError(f'Mapping {mapping_name} has non-string value for key {key}: {values}')


@dataclass
class Series:
    path: Path
    patient_id: str
    study_id: str

    # image metadata
    filenames: Optional[List[str]] = None
    spacing_inplane: Optional[Sequence[float]] = None
    metadata: Optional[Metadata] = field(default_factory=dict)

    mappings: List[str] = field(default_factory=list)
    error: Optional[Exception] = None
    _log: List[str] = field(default_factory=list)

    def __post_init__(self):
        if not self.path.exists():
            raise ArchiveItemPathNotFoundError(self.path)
        if not self.path.is_dir():
            raise NotADirectoryError(self.path)

    def extract_metadata(self, verify_dicom_filenames: bool = True) -> None:
        """
        Verify DICOM slices and extract metadata from the last DICOM slice
        """
        # try:
        reader = DICOMImageReader(self.path, verify_dicom_filenames=verify_dicom_filenames)
        self.metadata = reader.metadata
        self.spacing_inplane = self.metadata["spacing_inplane"]
        self.filenames = reader.dicom_slice_paths
        self.write_log('Extracted metadata')

    @staticmethod
    def metadata_matches(
        metadata: Metadata,
        mapping: Mapping,
        values_match_func: Callable[[str, str], bool],
    ) -> bool:
        """
        Determine whether Series' metadata matches the mapping.
        By default, values are trimmed from whitespace and case-insensitively compared.
        """
        for dicom_tag, allowed_values in mapping.items():
            dicom_tag = dicom_tag.lower().strip()
            if dicom_tag not in metadata:
                # metadata does not exist
                return False

            # check if the observed value matches with any of the allowed values
            if not any(values_match_func(needle=value, haystack=metadata[dicom_tag]) for value in allowed_values):
                return False

        return True

    def apply_mappings(
        self,
        mappings: Mappings,
        metadata_match_func: Optional[Callable[[Metadata, Mappings], bool]] = None,
        values_match_func: Optional[Union[Callable[[str, str], bool], str]] = "lower_strip_equals",
    ) -> None:
        """
        Apply mappings to the series
        """
        # resolve metadata match function
        if metadata_match_func is None:
            metadata_match_func = self.metadata_matches

        # resolve value match function
        if isinstance(values_match_func, str):
            variant = values_match_func

            def values_match_func(needle: str, haystack: str) -> bool:
                if "lower" in variant:
                    needle = needle.lower()
                    haystack = haystack.lower()
                if "strip" in variant:
                    needle = needle.strip()
                    haystack = haystack.strip()
                if "equals" in variant:
                    return needle == haystack
                elif "contains" in variant:
                    return needle in haystack
                elif "regex" in variant:
                    return re.search(needle, haystack) is not None
                else:
                    raise ValueError(f'Unknown values match function variant {variant}')

        for name, mapping in mappings.items():
            if metadata_match_func(metadata=self.metadata, mapping=mapping, values_match_func=values_match_func):
                self.mappings.append(name)

        if len(self.mappings) == 0:
            raise NoMappingsApplyError()
        self.write_log(f'Applied mappings [{", ".join(self.mappings)}]')

    def write_log(self, msg: str) -> None:
        self._log.append(msg)

    def compile_log(self) -> str:
        log = [f'\t{item}' for item in self._log]
        return '\n'.join([self.path.as_posix()] + log + [f'\tFATAL: {self.error}\n' if not self.is_valid else ''])

    @property
    def is_valid(self) -> bool:
        return self.error is None

    def __repr__(self) -> str:
        return f"Series({self.path.name})"


@dataclass
class _Dicom2MHACaseBase:
    input_dir: Path
    paths: List[PathLike]
    settings: Dicom2MHASettings


@dataclass
class Dicom2MHACase(Case, _Dicom2MHACaseBase):
    series: List[Series] = field(default_factory=list)

    def convert_item(self, output_dir: Path) -> None:
        self.initialize()
        self.extract_metadata()
        self.apply_mappings()
        self.resolve_duplicates()
        self.process_and_write(output_dir)

    def initialize(self) -> None:
        self.write_log(f'Importing {plural(len(self.paths), "serie")}')

        full_paths = set()
        for path in self.paths:
            full_path = self.input_dir / path
            serie = Series(full_path, self.patient_id, self.study_id)

            # if we find duplicate paths with the same patient and study id,
            # invalidate this series and continue for logging purposes
            if path in full_paths:
                raise FileExistsError(path)
            full_paths.add(full_path)
            self.write_log(f'\t+ ({len(self.series)}) {full_path}')
            self.series.append(serie)

        if not self.is_valid:
            self.invalidate()

    def extract_metadata(self) -> None:
        self.write_log(f'Extracting metadata from {plural(len(self.valid_series), "serie")}')
        errors = []

        for i, serie in enumerate(self.valid_series):
            try:
                serie.extract_metadata(
                    verify_dicom_filenames=self.settings.verify_dicom_filenames
                )
            except (MissingDICOMFilesError, UnreadableDICOMError) as e:
                serie.error = e
                errors.append(i)

        self.write_log(f'\t({plural(len(errors), "error")}{f" {errors}" if len(errors) > 0 else ""})')

    def apply_mappings(self) -> None:
        self.write_log(f'Applying mappings to {len(self.valid_series)} series')
        errors = []

        for i, serie in enumerate(self.valid_series):
            try:
                serie.apply_mappings(
                    mappings=self.settings.mappings,
                    metadata_match_func=self.settings.metadata_match_func,
                    values_match_func=self.settings.values_match_func,
                )
            except NoMappingsApplyError as e:
                serie.error = e
                errors.append(i)

        self.write_log(f'\t({plural(len(errors), "error")}{f" {errors}" if len(errors) > 0 else ""})')

    def resolve_duplicates(self) -> None:
        self.write_log(f'Resolving duplicates between {plural(len(self.valid_series), "serie")}')

        # define tiebreakers, which should have: name, value_func, pick_largest
        tiebreakers = [
            ('slice count', lambda a: len(a.filenames), True),
            ('image resolution', lambda a: np.prod(a.spacing_inplane), False),
            ('filename', lambda a: str(a.path), False),
        ]

        # create dict collecting all items for each mapping
        matched_series: Dict[str, List[Series]] = {
            mapping: [] for serie in self.valid_series for mapping in serie.mappings
        }
        for serie in self.valid_series:
            for mapping in serie.mappings:
                matched_series[mapping] += [serie]

        # use tiebreakers to select a single item for each mapping
        for mapping, series in matched_series.items():
            if self.settings.allow_duplicates:
                for i, serie in enumerate(series):
                    serie.mappings.remove(mapping)
                    serie.mappings.append(f'{mapping}_{i}')
            else:
                for name, value_func, pick_largest in tiebreakers:
                    if len(series) > 1:
                        serie_value_pairs = [(serie, value_func(serie)) for serie in series]
                        serie_value_pairs.sort(key=lambda a: a[1], reverse=pick_largest)
                        _, best_value = serie_value_pairs[0]

                        for serie, value in serie_value_pairs:
                            if value != best_value:
                                serie.mappings.remove(mapping)
                                serie.write_log(f'Removed by {name} tiebreaker from "{mapping}"')
                                series.remove(serie)

    def process_and_write(self, output_dir: Path) -> None:
        total = sum([len(serie.mappings) for serie in self.valid_series])
        self.write_log(f'Writing {plural(total, "serie")}')
        errors, skips = [], []

        patient_dir = output_dir / self.patient_id
        for i, serie in enumerate(self.valid_series):
            for mapping in serie.mappings:
                mapping_save_name = mapping
                if ":" in mapping_save_name:
                    mapping_save_name = mapping_save_name.split(':')[0]
                dst_path = patient_dir / f"{self.subject_id}_{mapping_save_name}.mha"
                if dst_path.exists():
                    serie.write_log(f'Skipped "{mapping}", already exists: {dst_path}')
                    skips.append(i)
                    continue

                try:
                    image = DICOMImageReader(serie.path, verify_dicom_filenames=self.settings.verify_dicom_filenames).image
                except Exception as e:
                    serie.write_log(
                        f'Skipped "{mapping}", reading DICOM sequence failed, maybe corrupt data? Error: {e}')
                    if self.settings.verbose >= 2:
                        logging.error(traceback.format_exc())
                    else:
                        logging.error(str(e))
                    errors.append(i)
                else:
                    if self.settings.scan_postprocess_func is not None:
                        image = self.settings.scan_postprocess_func(image)
                    try:
                        atomic_image_write(image=image, path=dst_path, mkdir=True)
                    except Exception as e:
                        serie.write_log(f'Skipped "{mapping}", write error: {e}')
                        logging.error(str(e))
                        errors.append(i)
                    else:
                        serie.write_log(f'Wrote image to {dst_path}')

        self.write_log(f'Wrote {total - len(errors) - len(skips)} MHA files to {patient_dir.as_posix()}\n'
                       f'\t({plural(len(errors), "error")}{f" {errors}" if len(errors) > 0 else ""}, '
                       f'{len(skips)} skipped{f" {skips}" if len(skips) > 0 else ""})')

    def invalidate(self, error: Exception = None) -> None:
        if error is None:
            error = CriticalErrorInSiblingError()
        for serie in self.valid_series:
            serie.error = error

    @property
    def subject_id(self) -> str:
        return f"{self.patient_id}_{self.study_id}"

    @property
    def is_valid(self) -> bool:
        return all([serie.is_valid for serie in self.series])

    @property
    def valid_series(self) -> List[Series]:
        return [item for item in self.series if item.is_valid]

    def write_log(self, msg: str) -> None:
        self._log.append(msg)

    def compile_log(self) -> str:
        """For questions: Stan.Noordman@Radboudumc.nl"""
        if self.settings.verbose == 0:
            return

        divider = '=' * 120
        summary = {}
        serie_log = []

        # summarize each serie's log (if any)
        for i, serie in enumerate(self.series):
            serie_log.append(f'({i}) {serie.compile_log()}')
            if serie.error:
                summary[serie.error.__class__.__name__] = summary.get(serie.error.__class__.__name__, []) + [i]

        # these are the errors that are not fatal
        ignored_errors = {e.__name__ for e in [NoMappingsApplyError]}

        # check if we should log any errors
        if len(set(summary.keys()).difference(ignored_errors)) > 0 or self.settings.verbose >= 2:
            # don't worry, this just looks nice in the log
            return '\n'.join([divider,
                              f'CASE {self.patient_id}_{self.study_id}',
                              f'\tPATIENT ID\t{self.patient_id}',
                              f'\tSTUDY ID\t{self.study_id}\n',
                              *self._log,
                              '\nSERIES', divider.replace('=', '-'),
                              'Errors found:',
                              *[f'\t{key}: {value}' for key, value in summary.items()],
                              '', *serie_log, ''])

    def cleanup(self) -> None:
        self.series = None
        super().cleanup()

    def __repr__(self) -> str:
        return f'Case({self.subject_id})'


class Dicom2MHAConverter(Converter):
    def __init__(
        self,
        input_dir: PathLike,
        output_dir: PathLike,
        dcm2mha_settings: Union[PathLike, Dict] = None,
        case_class: Case = Dicom2MHACase,
    ):
        """
        Convert DICOM Archive to MHA Archive.
        See https://github.com/DIAGNijmegen/picai_prep for additional documentation.

        Parameters
        ----------
        input_dir: PathLike
            path to the DICOM archive. Used as base path for the relative paths of the archive items.
        output_dir: PathLike
            path to store the resulting MHA archive.
        dcm2mha_settings: Union[PathLike, Dict], default: None
            object with mappings, cases and optional parameters. May be a dictionary containing mappings, archive,
            and optionally options, or a path to a JSON file with these elements.
            - mappings: criteria to map DICOM sequences to their MHA counterparts
            - cases: list of DICOM sequences in the DICOM archive. Each case is to be an object with a patient_id,
                study_id and path to DICOM sequence
            - options: (optional)
                - num_threads: number of multithreading threads.
                    Default: 4.
                - verify_dicom_filenames: whether to check if DICOM filenames contain consecutive
                    numbers.
                    Default: True
                - allow_duplicates: whether multiple DICOM series can map to the same MHA postfix.
                    Default: False
                - metadata_match_func: method to match DICOM metadata to MHA sequences. Only use
                    this if you know what you're doing.
                    Default: check if the observed value matches with any of the allowed values.
                - values_match_func: criteria to consider two values a match, when comparing the
                    value from the DICOM metadata against the provided allowed vaues in the mapping.
                    Default: lower_strip_equals (case-insensitive matching of values without trailing or leading spaces)
                - verbose: control logfile verbosity. 0 does not output a logfile,
                    1 logs cases which have critically failed, 2 logs all cases (may lead to
                    very large log files)
                    Default: 1
        """
        self.input_dir = Path(input_dir)
        self.output_dir = Path(output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)
        self.case_class = case_class

        # parse settings
        if isinstance(dcm2mha_settings, (Path, str)):
            with open(dcm2mha_settings) as fp:
                dcm2mha_settings = json.load(fp)

        jsonschema.validate(dcm2mha_settings, dcm2mha_schema, cls=jsonschema.Draft7Validator)
        self.settings = Dicom2MHASettings(
            mappings=dcm2mha_settings['mappings'],
            **dcm2mha_settings.get('options', {})
        )

        self.cases = self._init_cases(dcm2mha_settings['archive'])

        self.initialize_log(self.output_dir, self.settings.verbose)

    def _init_cases(self, archive: List[Dict]) -> List[Case]:
        cases = {}
        for item in archive:
            key = tuple(item[id] for id in ["patient_id", "study_id"])
            cases[key] = cases.get(key, []) + [item['path']]
        return [
            self.case_class(input_dir=self.input_dir, patient_id=patient_id,
                            study_id=study_id, paths=paths, settings=self.settings)
            for (patient_id, study_id), paths in cases.items()
        ]

    def convert(self) -> None:
        self._convert(
            title='Dicom2MHA',
            cases=self.cases,
            parameters={
                'output_dir': self.output_dir
            },
            num_threads=self.settings.num_threads,
        )


class DICOMImageReader:
    """
    Read folder containing DICOM slices (possibly enclosed in a 'dicom.zip' file).
    If both DICOM slices and a 'dicom.zip' file are present, the dicom.zip used.

    Parameters
    ----------
    image_series_path: PathLike
        path to the folder containing the DICOM slices. The folder should contain the DICOM slices,
        or a zip file named "dicom.zip" containing the DICOM slices.

    Usage
    -----
    >>> image = DICOMImageReader('path/to/dicom/folder').image

    or

    >>> reader = DICOMImageReader('path/to/dicom/folder')
    >>> image = reader.image
    >>> metadata = reader.metadata
    """

    def __init__(self, path: PathLike, verify_dicom_filenames: bool = True):
        self.path = Path(path)
        self.verify_dicom_filenames = verify_dicom_filenames
        self._image = None
        self._metadata = None
        self.dicom_slice_paths: Optional[List[str]] = None

        self.series_reader = sitk.ImageSeriesReader()
        if (self.path / "dicom.zip").exists():
            self.path = self.path / "dicom.zip"
            with zipfile.ZipFile(self.path, "r") as zf:
                self.dicom_slice_paths = [
                    self.path / name
                    for name in zf.namelist()
                    if name.endswith(".dcm")
                ]
            if self.verify_dicom_filenames:
                self._verify_dicom_filenames()
        else:
            self._set_dicom_list()

    @property
    def image(self) -> sitk.Image:
        if self._image is None:
            self._image = self._read_image()
        return self._image

    @property
    def metadata(self) -> Dict[str, str]:
        if self._metadata is None:
            self._metadata = self._read_metadata()
        return self._metadata

    def _read_image(self, path: Optional[PathLike] = None) -> sitk.Image:
        if path is None:
            path = self.path
        path = Path(path)

        if path.name == "dicom.zip":
            return self._read_image_dicom_zip(path=path)
        try:
            return self._read_image_sitk(path=path)
        except RuntimeError:
            # try again with pydicom
            return self._read_image_pydicom(path=path)

    @staticmethod
    def _filter_localizer_slices(dicom_slice_paths: List[str]) -> List[str]:
        """
        Filter out localizer slices (slices with ImageType == LOCALIZER).
        WARNING: this is slow and a heuristic that may not work for all datasets.
        """
        filtered_dicom_slice_paths = []
        for path in dicom_slice_paths:
            reader = sitk.ImageFileReader()
            reader.SetFileName(str(path))
            reader.LoadPrivateTagsOn()
            reader.ReadImageInformation()
            image_type = reader.GetMetaData("0008|0008")
            if "LOCALIZER" not in image_type.upper():
                filtered_dicom_slice_paths.append(path)
        return filtered_dicom_slice_paths

    def _set_dicom_list(self, path: Optional[PathLike] = None) -> None:
        """
        Set the list of paths to the DICOM slices.

        Parameters
        ----------
        path: PathLike
            path to the folder containing the DICOM slices. The folder should contain the DICOM slices,
            or a zip file named "dicom.zip" containing the DICOM slices.
            default: self.path
        """
        if path is None:
            path = self.path

        self.dicom_slice_paths = self.series_reader.GetGDCMSeriesFileNames(str(path))

        # verify DICOM files are found
        if len(self.dicom_slice_paths) == 0:
            raise MissingDICOMFilesError(self.path)

        if self.verify_dicom_filenames:
            self._verify_dicom_filenames()

    def _read_image_sitk(
        self,
        path: Optional[PathLike] = None,
    ) -> sitk.Image:
        """
        Read image using SimpleITK.

        Parameters
        ----------
        path: PathLike
            path to the folder containing the DICOM slices. The folder should contain the DICOM slices,
            or a zip file named "dicom.zip" containing the DICOM slices.
            default: self.path

        Returns
        -------
        image: SimpleITK.Image
        """
        if path is not None:
            self.path = path
            self._set_dicom_list(path=path)

        # read DICOM sequence
        try:
            self.series_reader.SetFileNames(self.dicom_slice_paths)
            image: sitk.Image = self.series_reader.Execute()
        except RuntimeError:
            # try again while removing localizer slices
            self.dicom_slice_paths = self._filter_localizer_slices(self.dicom_slice_paths)
            self.series_reader.SetFileNames(self.dicom_slice_paths)
            image: sitk.Image = self.series_reader.Execute()

        # read metadata from the last DICOM slice
        reader = sitk.ImageFileReader()
        reader.SetFileName(self.dicom_slice_paths[0])
        reader.LoadPrivateTagsOn()
        reader.ReadImageInformation()

        # set metadata
        metadata = {key: reader.GetMetaData(key).strip() for key in reader.GetMetaDataKeys()}
        for key, value in metadata.items():
            if len(value) > 0:
                image.SetMetaData(key, value)

        return image

    def _read_image_pydicom(self, path: Optional[PathLike] = None) -> sitk.Image:
        """
        Read image using pydicom. Warning: experimental! This function has limited capabilities.

        Parameters
        ----------
        path: PathLike
            path to the folder containing the DICOM slices. The folder should contain the DICOM slices,
            or a zip file named "dicom.zip" containing the DICOM slices.
            default: self.path

        Returns
        -------
        image: SimpleITK.Image
        """
        if path is not None:
            self.path = path
            self._set_dicom_list(path=path)

        files = [pydicom.dcmread(dcm) for dcm in self.dicom_slice_paths]

        # skip files with no SliceLocation (eg. scout views)
        slices = filter(lambda a: hasattr(a, 'SliceLocation'), files)
        slices = sorted(slices, key=lambda s: s.SliceLocation)

        # create and fill 3D array
        image = np.zeros([len(slices)] + list(slices[0].pixel_array.shape))
        for i, s in enumerate(slices):
            image[i, :, :] = s.pixel_array

        # convert to SimpleITK
        image: sitk.Image = sitk.GetImageFromArray(image)
        ref = slices[0]  # corresponds to the same slice as with SimpleITK
        image.SetSpacing(list(ref.PixelSpacing) + [ref.SliceThickness])
        image.SetOrigin(ref.ImagePositionPatient)
        image.SetDirection(self.get_orientation_tuple_sitk(ref))

        for key in ref.keys():
            # collect all available metadata (with DICOM tags, e.g. 0010|1010, as keys)
            value = self.get_pydicom_value(ref, key)
            if value is not None:
                key = str(key).replace(",", "|").replace("(", "").replace(")", "").replace(" ", "")
                image.SetMetaData(key, value)

        return image

    def _read_image_dicom_zip(self, path: PathLike) -> sitk.Image:
        """
        Read image from dicom.zip file using SimpleITK.
        """
        with zipfile.ZipFile(path) as zf:
            if not zf.namelist():
                raise RuntimeError('dicom.zip is empty')

            with tempfile.TemporaryDirectory() as tempdir:
                zf.extractall(tempdir)
                return self._read_image(tempdir)

    def _read_metadata(self) -> Dict[str, str]:
        if self._image is not None:
            return self._collect_metadata_sitk(self._image)

        if self.path.name == "dicom.zip":
            # read metadata from dicom.zip with pydicom
            with zipfile.ZipFile(self.path) as zf:
                if not zf.namelist():
                    raise RuntimeError('dicom.zip is empty')

                with tempfile.TemporaryDirectory() as tempdir:
                    targetpath = zf.extract(member=zf.namelist()[-1], path=tempdir)
                    return self._read_metadata_from_file(targetpath)

        # extract metadata from first/last(?) DICOM slice
        dicom_slice_path = self.dicom_slice_paths[0]
        return self._read_metadata_from_file(dicom_slice_path)

    def _read_metadata_from_file(self, path: PathLike) -> Dict[str, str]:
        try:
            reader = sitk.ImageFileReader()
            reader.SetFileName(str(path))
            reader.LoadPrivateTagsOn()
            reader.ReadImageInformation()
            return self._collect_metadata_sitk(reader)
        except Exception:
            try:
                with pydicom.dcmread(path, stop_before_pixels=True) as ds:
                    return self._collect_metadata_pydicom(ds)
            except pydicom.errors.InvalidDicomError:
                raise UnreadableDICOMError(path)

    def _collect_metadata_sitk(self, ref: Union[sitk.Image, sitk.ImageFileReader]) -> Dict[str, str]:
        metadata = {}
        for key in ref.GetMetaDataKeys():
            # collect all available metadata (with DICOM tags, e.g. 0010|1010, as keys)
            metadata[key] = ref.GetMetaData(key).strip()
        for name, key in dicom_tags.items():
            # collect metadata with DICOM names, e.g. patientsage, as keys)
            metadata[name] = ref.GetMetaData(key).strip() if ref.HasMetaDataKey(key) else ''

        metadata["spacing_inplane"] = ref.GetSpacing()[0:2]
        metadata["image_direction"] = ref.GetDirection()
        return metadata

    def _collect_metadata_pydicom(self, ds: "pydicom.dataset.Dataset") -> Dict[str, str]:
        metadata = {}
        for key in ds.keys():
            # collect all available metadata (with DICOM tags, e.g. 0010|1010, as keys)
            value = self.get_pydicom_value(ds, key)
            if value is not None:
                key = str(key).replace(", ", "|").replace("(", "").replace(")", "")
                metadata[key] = value
        for key in dicom_tags.values():
            # collect metadata with DICOM names, e.g. patientsage, as keys)
            value = self.get_pydicom_value(ds, key)
            metadata[key] = value if value is not None else ''

        metadata["spacing_inplane"] = ds.PixelSpacing[0:2]
        return metadata

    @staticmethod
    def get_pydicom_value(ds: pydicom.dataset.Dataset, key: "Union[pydicom.tag.BaseTag, str]") -> str:
        if isinstance(key, str):
            key = '0x' + key.replace('|', '')
        if key in ds:
            result = ds[key]
            if result.is_empty:
                return ''
            result = result.value
            if isinstance(result, (list, pydicom.multival.MultiValue)):
                result = "\\".join([str(v) for v in result])
            return str(result)
        return ''

    @staticmethod
    def get_orientation_matrix(ds: pydicom.FileDataset) -> np.ndarray:
        x, y = np.array(list(map(float, ds.ImageOrientationPatient))).reshape(2, 3)
        return np.stack([x, y, np.cross(x, y)])

    def get_orientation_tuple_sitk(self, ds: pydicom.FileDataset) -> Tuple:
        return tuple(self.get_orientation_matrix(ds).transpose().flatten())

    def _verify_dicom_filenames(self, filenames: Optional[List[PathLike]] = None) -> bool:
        """
        Verify DICOM filenames have increasing numbers, with no gaps

        Common prefixes are removed from the filenames before checking the numbers,
        this allows to verify filenames like "1.2.86.1.dcm", ..., "1.2.86.12.dcm".
        """
        if filenames is None:
            filenames = [os.path.basename(dcm) for dcm in self.dicom_slice_paths]

        # remove common prefixes
        common_prefix = os.path.commonprefix(filenames)
        if common_prefix:
            filenames = [fn.replace(common_prefix, "") for fn in filenames]
        common_postfix = os.path.commonprefix([fn[::-1] for fn in filenames])[::-1]
        if common_postfix:
            filenames = [fn.replace(common_postfix, "") for fn in filenames]

        # extract numbers from filenames
        filename_digits = [(''.join(c for c in str(fn) if c.isdigit())) for fn in filenames]
        filename_digits = [int(d) for d in filename_digits if d]
        if len(filename_digits) < 2:
            # either no numbers in the filenames, or only one file
            return True

        missing_slices = False
        for num in range(min(filename_digits), max(filename_digits) + 1):
            if num not in filename_digits:
                missing_slices = True
                break
        if missing_slices:
            raise MissingDICOMFilesError(self.path)
        return True

    def __repr__(self) -> str:
        return f'DICOMImageReader({self.path})'



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import json
import os
from typing import Optional

from picai_prep.data_utils import PathLike
from tqdm import tqdm


def generate_mha2nnunet_settings(
    archive_dir: PathLike,
    output_path: PathLike,
    annotations_dir: Optional[PathLike] = None,
    task: str = "Task2201_picai_baseline",
):
    """
    Create mha2nnunet_settings.json (for inference) for an MHA archive with the following structure:
    /path/to/archive/
    ├── [patient UID]/
        ├── [patient UID]_[study UID]_[modality].mha
        ...

    Parameters:
    - archive_dir: path to MHA archive
    - output_path: path to store MHA -> nnUNet settings JSON to
        (parent folder should exist)
    """
    archive_list = []

    # traverse MHA archive
    for patient_id in tqdm(sorted(os.listdir(archive_dir))):
        # traverse each patient's studies
        patient_dir = os.path.join(archive_dir, patient_id)
        if not os.path.isdir(patient_dir):
            continue

        # collect list of available studies
        files = os.listdir(patient_dir)
        files = [fn.replace(".mha", "") for fn in files if ".mha" in fn and "._" not in fn]
        subject_ids = ["_".join(fn.split("_")[0:2]) for fn in files]
        subject_ids = sorted(list(set(subject_ids)))

        # check which studies are complete
        for subject_id in subject_ids:
            patient_id, study_id = subject_id.split("_")

            # construct scan paths
            scan_paths = [
                f"{patient_id}/{subject_id}_{modality}.mha"
                for modality in ["t2w", "adc", "hbv"]
            ]
            all_scans_found = all([
                os.path.exists(os.path.join(archive_dir, path))
                for path in scan_paths
            ])

            # construct annotation path
            annotation_path = f"{subject_id}.nii.gz"

            if annotations_dir is not None:
                # check if annotation exists
                if not os.path.exists(os.path.join(annotations_dir, annotation_path)):
                    # could not find annotation, skip case
                    continue

            if all_scans_found:
                # store info for complete studies
                archive_list += [{
                    "patient_id": patient_id,
                    "study_id": study_id,
                    "scan_paths": scan_paths,
                    "annotation_path": annotation_path,
                }]

    mha2nnunet_settings = {
        "dataset_json": {
            "task": task,
            "description": "bpMRI scans from PI-CAI dataset to train nnUNet baseline",
            "tensorImageSize": "4D",
            "reference": "",
            "licence": "",
            "release": "1.0",
            "modality": {
                "0": "T2W",
                "1": "CT",
                "2": "HBV"
            },
            "labels": {
                "0": "background",
                "1": "lesion"
            }
        },
        "preprocessing": {
            # optionally, resample and perform centre crop:
            # "matrix_size": [
            #     20,
            #     320,
            #     320
            # ],
            # "spacing": [
            #     3.0,
            #     0.5,
            #     0.5
            # ],
        },
        "archive": archive_list
    }

    if not len(archive_list):
        raise ValueError(f"Did not find any MHA scans in {archive_dir}, aborting.")

    with open(output_path, "w") as fp:
        json.dump(mha2nnunet_settings, fp, indent=4)

    print(f""""
    Saved mha2nnunet_settings to {output_path}, with {len(archive_list)} cases.
    """)



from picai_prep.data_utils import PathLike


class ConverterException(Exception):
    """Base Exception for errors in an item (series within a case)"""

    def __init__(self, message: str):
        super().__init__(message)

    def __str__(self):
        return f'{type(self).__name__}: {", ".join([a for a in self.args])}'


class ArchiveItemPathNotFoundError(ConverterException):
    """Exception raised when a archive path could not be found"""

    def __init__(self, path: PathLike):
        super().__init__(f"Provided archive item path not found ({path})")


class CriticalErrorInSiblingError(ConverterException):
    """Exception raised when a critical error in a sibling item occurs"""

    def __init__(self):
        super().__init__("Critical error in sibling item")


class MissingDICOMFilesError(ConverterException):
    """Exception raised when a DICOM series has missing DICOM slices"""

    def __init__(self, path: PathLike):
        super().__init__(f"Missing DICOM slices detected in {path}")


class NoMappingsApplyError(ConverterException):
    """Exception raised when no mappings apply to the case"""

    def __init__(self):
        super().__init__('None of the provided mappings apply to this item')


class UnreadableDICOMError(ConverterException):
    """Exception raised when a DICOM series could not be loaded"""

    def __init__(self, path: PathLike):
        super().__init__(f'Could not read {path} using either SimpleITK or pydicom')


class DCESeriesNotFoundError(ConverterException):
    """Exception raised when DCE series could not be found"""

    def __init__(self, subject_id: str):
        super().__init__(f"DCE series not found for {subject_id}")



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from dataclasses import dataclass
from typing import Any, Callable, Iterable, List, Optional, Tuple, Union

import numpy as np
import SimpleITK as sitk
from numpy.testing import assert_allclose
from scipy import ndimage

try:
    import numpy.typing as npt
except ImportError:  # pragma: no cover
    pass


@dataclass
class PreprocessingSettings():
    """
    Preprocessing settings
    - matrix_size: number of voxels output image (z, y, x)
    - spacing: output voxel spacing in mm/voxel (z, y, x)
    - physical_size: size in mm of the target image (z, y, x)
    - crop_only: only crop to specified size (i.e., do not pad)
    - align_segmentation: whether to align the scans using the centroid of the provided segmentation
    - scan_interpolator: interpolation method for scans
    - lbl_interpolator: interpolation method for labels
    - seg_interpolator: interpolation method for segmentations
    """
    matrix_size: Optional[Iterable[int]] = None
    spacing: Optional[Iterable[float]] = None
    physical_size: Optional[Iterable[float]] = None
    crop_only: bool = False
    align_segmentation: Optional[sitk.Image] = None
    scan_interpolator: int = sitk.sitkBSpline
    lbl_interpolator: int = sitk.sitkNearestNeighbor
    seg_interpolator: int = sitk.sitkNearestNeighbor

    def __post_init__(self):
        if self.physical_size is None and self.spacing is not None and self.matrix_size is not None:
            # calculate physical size
            self.physical_size = [
                voxel_spacing * num_voxels
                for voxel_spacing, num_voxels in zip(
                    self.spacing,
                    self.matrix_size
                )
            ]

        if self.spacing is None and self.physical_size is not None and self.matrix_size is not None:
            # calculate spacing
            self.spacing = [
                size / num_voxels
                for size, num_voxels in zip(
                    self.physical_size,
                    self.matrix_size
                )
            ]

        if self.align_segmentation is not None:
            raise NotImplementedError("Alignment of scans based on segmentation is not implemented yet.")


def resample_img(
    image: sitk.Image,
    out_spacing: Iterable[float] = (2.0, 2.0, 2.0),
    out_size: Optional[Iterable[int]] = None,
    is_label: bool = False,
    interpolation = None,
    pad_value: Optional[Union[float, int]] = 0.,
) -> sitk.Image:
    """
    Resample images to target resolution spacing
    Ref: SimpleITK
    """
    # get original spacing and size
    original_spacing = image.GetSpacing()
    original_size = image.GetSize()

    # convert our z, y, x convention to SimpleITK's convention
    out_spacing = list(out_spacing)[::-1]

    if out_size is None:
        # calculate output size in voxels
        out_size = [
            int(np.round(
                size * (spacing_in / spacing_out)
            ))
            for size, spacing_in, spacing_out in zip(original_size, original_spacing, out_spacing)
        ]

    # determine pad value
    if pad_value is None:
        pad_value = image.GetPixelIDValue()

    # set up resampler
    resample = sitk.ResampleImageFilter()
    resample.SetOutputSpacing(list(out_spacing))
    resample.SetSize(out_size)
    resample.SetOutputDirection(image.GetDirection())
    resample.SetOutputOrigin(image.GetOrigin())
    resample.SetTransform(sitk.Transform())
    resample.SetDefaultPixelValue(pad_value)
    if interpolation is not None:
        resample.SetInterpolator(interpolation)
    elif is_label:
        resample.SetInterpolator(sitk.sitkNearestNeighbor)
    else:
        resample.SetInterpolator(sitk.sitkBSpline)

    # perform resampling
    image = resample.Execute(image)

    return image


def input_verification_crop_or_pad(
    image: "Union[sitk.Image, npt.NDArray[Any]]",
    size: Optional[Iterable[int]] = (20, 256, 256),
    physical_size: Optional[Iterable[float]] = None,
) -> Tuple[Iterable[int], Iterable[int]]:
    """
    Calculate target size for cropping and/or padding input image

    Parameters:
    - image: image to be resized (sitk.Image or numpy.ndarray)
    - size: target size in voxels (z, y, x)
    - physical_size: target size in mm (z, y, x)

    Either size or physical_size must be provided.

    Returns:
    - shape of original image (in convention of SimpleITK (x, y, z) or numpy (z, y, x))
    - size of target image (in convention of SimpleITK (x, y, z) or numpy (z, y, x))
    """
    # input conversion and verification
    if physical_size is not None:
        # convert physical size to voxel size (only supported for SimpleITK)
        if not isinstance(image, sitk.Image):
            raise ValueError("Crop/padding by physical size is only supported for SimpleITK images.")
        spacing_zyx = list(image.GetSpacing())[::-1]
        size_zyx = [length/spacing for length, spacing in zip(physical_size, spacing_zyx)]
        size_zyx = [int(np.round(x)) for x in size_zyx]

        if size is None:
            # use physical size
            size = size_zyx
        else:
            # verify size
            if list(size) != list(size_zyx):
                raise ValueError(f"Size and physical size do not match. Size: {size}, physical size: "
                                 f"{physical_size}, spacing: {spacing_zyx}, size_zyx: {size_zyx}.")

    if isinstance(image, sitk.Image):
        # determine shape and convert convention of (z, y, x) to (x, y, z) for SimpleITK
        shape = image.GetSize()
        size = list(size)[::-1]
    else:
        # determine shape for numpy array
        assert isinstance(image, (np.ndarray, np.generic))
        shape = image.shape
        size = list(size)
    rank = len(size)
    assert rank <= len(shape) <= rank + 1, \
        f"Example size doesn't fit image size. Got shape={shape}, output size={size}"

    return shape, size


def crop_or_pad(
    image: "Union[sitk.Image, npt.NDArray[Any]]",
    size: Optional[Iterable[int]] = (20, 256, 256),
    physical_size: Optional[Iterable[float]] = None,
    crop_only: bool = False,
    pad_only: bool = False,
) -> "Union[sitk.Image, npt.NDArray[Any]]":
    """
    Resize image by cropping and/or padding

    Parameters:
    - image: image to be resized (sitk.Image or numpy.ndarray)
    - size: target size in voxels (z, y, x)
    - physical_size: target size in mm (z, y, x)

    Either size or physical_size must be provided.

    Returns:
    - resized image (same type as input)
    """
    # input conversion and verification
    shape, size = input_verification_crop_or_pad(image, size, physical_size)

    # set identity operations for cropping and padding
    rank = len(size)
    padding = [[0, 0] for _ in range(rank)]
    slicer = [slice(None) for _ in range(rank)]

    # for each dimension, determine process (cropping or padding)
    for i in range(rank):
        if shape[i] < size[i]:
            if crop_only:
                continue

            # set padding settings
            padding[i][0] = (size[i] - shape[i]) // 2
            padding[i][1] = size[i] - shape[i] - padding[i][0]
        else:
            if pad_only:
                continue

            # create slicer object to crop image
            idx_start = int(np.floor((shape[i] - size[i]) / 2.))
            idx_end = idx_start + size[i]
            slicer[i] = slice(idx_start, idx_end)

    # crop and/or pad image
    if isinstance(image, sitk.Image):
        pad_filter = sitk.ConstantPadImageFilter()
        pad_filter.SetPadLowerBound([pad[0] for pad in padding])
        pad_filter.SetPadUpperBound([pad[1] for pad in padding])
        return pad_filter.Execute(image[tuple(slicer)])
    else:
        return np.pad(image[tuple(slicer)], padding)


@dataclass
class Sample:
    scans: List[sitk.Image]
    lbl: Optional[sitk.Image] = None
    seg: Optional[sitk.Image] = None
    name: Optional[str] = None
    settings: PreprocessingSettings = None
    lbl_preprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    lbl_postprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    seg_preprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    seg_postprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    scan_preprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    scan_postprocess_func: Optional[Callable[[sitk.Image], sitk.Image]] = None
    case_preprocess_func: Optional[Callable[["Sample"], "Sample"]] = None
    case_postprocess_func: Optional[Callable[["Sample"], "Sample"]] = None
    num_gt_lesions: Optional[int] = None

    def __post_init__(self):
        if self.lbl is not None:
            # keep track of connected components
            lbl = sitk.GetArrayFromImage(self.lbl)
            _, num_gt_lesions = ndimage.label(lbl, structure=np.ones((3, 3, 3)))
            self.num_gt_lesions = num_gt_lesions

        if self.settings is None:
            self.settings = PreprocessingSettings()

    def resample_to_first_scan(self):
        """Resample scans and label to the first scan"""
        # set up resampler to resolution, field of view, etc. of first scan
        resampler = sitk.ResampleImageFilter()  # default linear
        resampler.SetReferenceImage(self.scans[0])
        resampler.SetInterpolator(self.settings.scan_interpolator)

        # resample other images
        self.scans[1:] = [resampler.Execute(scan) for scan in self.scans[1:]]

        # resample annotation and segmentation
        resampler.SetInterpolator(self.settings.lbl_interpolator)
        if self.lbl is not None:
            self.lbl = resampler.Execute(self.lbl)
        if self.seg is not None:
            self.seg = resampler.Execute(self.seg)

    def resample_spacing(self, spacing: Optional[Iterable[float]] = None):
        """Resample scans and label to the target spacing"""
        if spacing is None:
            assert self.settings.spacing is not None
            spacing = self.settings.spacing

        # resample scans to target resolution
        self.scans = [
            resample_img(scan, out_spacing=spacing, interpolation=self.settings.scan_interpolator)
            for scan in self.scans
        ]

        # resample annotation to target resolution
        if self.lbl is not None:
            self.lbl = resample_img(self.lbl, out_spacing=spacing, interpolation=self.settings.lbl_interpolator)
        if self.seg is not None:
            self.seg = resample_img(self.seg, out_spacing=spacing, interpolation=self.settings.seg_interpolator)

    def centre_crop_or_pad(self):
        """Centre crop and/or pad scans and label"""
        kwargs = {
            "size": self.settings.matrix_size,
            "physical_size": self.settings.physical_size,
            "crop_only": self.settings.crop_only,
        }
        self.scans = [
            crop_or_pad(scan, **kwargs)
            for scan in self.scans
        ]

        if self.lbl is not None:
            self.lbl = crop_or_pad(self.lbl, **kwargs)
        if self.seg is not None:
            self.seg = crop_or_pad(self.seg, **kwargs)

    def align_physical_metadata(self, check_almost_equal=True):
        """Align the origin and direction of each scan, and label"""
        case_origin, case_direction, case_spacing = None, None, None
        for img in self.scans:
            # copy metadata of first scan (nnUNet and nnDetection require this to match exactly)
            if case_origin is None:
                case_origin = img.GetOrigin()
                case_direction = img.GetDirection()
                case_spacing = img.GetSpacing()
            else:
                if check_almost_equal:
                    # check if current scan's metadata is almost equal to the first scan
                    assert_allclose(img.GetOrigin(), case_origin)
                    assert_allclose(img.GetDirection(), case_direction)
                    assert_allclose(img.GetSpacing(), case_spacing)

                # copy over first scan's metadata to current scan
                img.SetOrigin(case_origin)
                img.SetDirection(case_direction)
                img.SetSpacing(case_spacing)

        if self.lbl is not None:
            assert case_origin is not None and case_direction is not None and case_spacing is not None
            self.lbl.SetOrigin(case_origin)
            self.lbl.SetDirection(case_direction)
            self.lbl.SetSpacing(case_spacing)
        if self.seg is not None:
            assert case_origin is not None and case_direction is not None and case_spacing is not None
            self.seg.SetOrigin(case_origin)
            self.seg.SetDirection(case_direction)
            self.seg.SetSpacing(case_spacing)

    def preprocess(self):
        """Perform all preprocessing steps"""
        # user-defined preprocessing steps
        if self.case_preprocess_func:
            # apply case transformation
            self = self.case_preprocess_func(self)
        if self.lbl is not None and self.lbl_preprocess_func:
            # apply label transformation
            self.lbl = self.lbl_preprocess_func(self.lbl)
        if self.seg is not None and self.seg_preprocess_func:
            # apply segmentation transformation
            self.seg = self.seg_preprocess_func(self.seg)
        if self.scan_preprocess_func:
            # apply scan transformation
            self.scans = [self.scan_preprocess_func(scan) for scan in self.scans]

        if self.settings.spacing is not None:
            # resample scans and label to specified spacing
            self.resample_spacing()

        if self.settings.matrix_size is not None or self.settings.physical_size is not None:
            # perform centre crop and/or pad
            self.centre_crop_or_pad()

        # resample scans and label to first scan's spacing, field-of-view, etc.
        self.resample_to_first_scan()

        # copy physical metadata to align subvoxel differences between sequences
        self.align_physical_metadata()

        if self.lbl is not None:
            # check connected components of annotation
            lbl = sitk.GetArrayFromImage(self.lbl)
            _, num_gt_lesions = ndimage.label(lbl, structure=np.ones((3, 3, 3)))
            assert self.num_gt_lesions == num_gt_lesions, \
                f"Label has changed due to resampling/other errors for {self.name}! " \
                + f"Have {self.num_gt_lesions} -> {num_gt_lesions} isolated ground truth lesions"

        # user-defined postprocessing steps
        if self.lbl is not None and self.lbl_postprocess_func:
            # apply label transformation
            self.lbl = self.lbl_postprocess_func(self.lbl)
        if self.seg is not None and self.seg_postprocess_func:
            # apply segmentation transformation
            self.seg = self.seg_postprocess_func(self.seg)
        if self.scan_postprocess_func:
            # apply scan transformation
            self.scans = [self.scan_postprocess_func(scan) for scan in self.scans]
        if self.case_postprocess_func:
            # apply case transformation
            self = self.case_postprocess_func(self)


def resample_to_reference_scan(
    image: "Union[npt.NDArray[Any], sitk.Image]",
    reference_scan_original: sitk.Image,
    reference_scan_preprocessed: Optional[sitk.Image] = None,
    interpolator: Optional[sitk.ResampleImageFilter] = None,
) -> sitk.Image:
    """
    Translate image/prediction/annotation to physical space of original scan (e.g., T2-weighted scan)

    Parameters:
    - image: scan, detection map or (softmax) prediction
    - reference_scan_original: SimpleITK image to which the prediction should be resampled and resized
    - reference_scan_preprocessed: (Optional) SimpleITK image with physical metadata for `image`
        (e.g., scan in nnUNet Raw Data Archive). If not provided, `image` should be a SimpleITK image.

    Returns:
    - resampled image, in same physical space as reference_scan_original
    """
    # convert image to SimpleITK image and copy physical metadata
    if not isinstance(image, sitk.Image):
        if reference_scan_preprocessed is None:
            raise ValueError("Need SimpleITK Image or reference scan for phsyical metadata!")
        image: sitk.Image = sitk.GetImageFromArray(image)

    if reference_scan_preprocessed is not None:
        image.CopyInformation(reference_scan_preprocessed)

    if interpolator is None:
        # determine interpolation method based on image dtype
        dtype_name = image.GetPixelIDTypeAsString()
        if "integer" in dtype_name:
            interpolator = sitk.sitkNearestNeighbor
        elif "float" in dtype_name:
            interpolator = sitk.sitkLinear
        else:
            raise ValueError(f"Unknown pixel type {dtype_name}")

    # prepare resampling to original scan
    resampler = sitk.ResampleImageFilter()  # default linear
    resampler.SetReferenceImage(reference_scan_original)
    resampler.SetInterpolator(interpolator)

    # resample image to original scan
    image = resampler.Execute(image)

    return image



import gc
import logging
import traceback
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

from tqdm import tqdm


@dataclass
class Case(ABC):
    patient_id: str
    study_id: str
    error: Optional[Exception] = None
    skip_conversion: bool = False
    _log: List[str] = field(default_factory=list)

    @abstractmethod
    def convert_item(self, **kwargs):
        """"
        Execute conversion process, please implement this.
        """
        raise NotImplementedError()

    def convert(self, **kwargs):
        """"
        Execute conversion process, while handling errors.
        Please override the `convert_item` method to implement the conversion process.
        """
        try:
            self.convert_item(**kwargs)
        except Exception as e:
            self.invalidate(e)
        self.compile_log()

    @property
    def subject_id(self):
        return f'{self.patient_id}_{self.study_id}'

    @property
    def is_valid(self):
        return self.error is None

    def invalidate(self, error: Exception):
        self.error = error
        self.error_trace = traceback.format_exc()

    def write_log(self, msg: str):
        self._log.append(msg)

    @abstractmethod
    def compile_log(self):
        raise NotImplementedError()

    def cleanup(self):
        self._log = None
        gc.collect()

    def __repr__(self):
        return f'Case({self.subject_id})'


class Converter:
    @staticmethod
    def initialize_log(output_dir: Path, verbose: int):
        if verbose >= 1:
            logfile = output_dir / f'picai_prep_{datetime.now().strftime("%Y%m%d%H%M%S")}.log'
            print(f'Writing log to {logfile.absolute()}')
            logging.basicConfig(level=logging.INFO, format='%(message)s', filename=logfile)
            logging.info(f'Output directory set to {output_dir.absolute().as_posix()}')
        else:
            logging.disable(logging.INFO)

    @staticmethod
    def _convert(title: str, cases: List[Case], parameters: Dict[str, Any], num_threads: int = 4):
        start_time = datetime.now()
        logging.info(f'{title} conversion started at {start_time.isoformat()}\n')
        num_cases_skipped = 0

        if num_threads >= 2:
            with ThreadPoolExecutor(max_workers=num_threads) as pool:
                futures = {pool.submit(case.convert, **parameters): case for case in cases}
                for future in tqdm(as_completed(futures), total=len(cases)):
                    case = futures[future]
                    case_log = future.result()
                    if case_log:
                        logging.info(case_log)
                    if case.skip_conversion:
                        num_cases_skipped += 1
                    case.cleanup()
        else:
            for case in tqdm(cases):
                case_log = case.convert(**parameters)
                if case_log:
                    logging.info(case_log)
                if case.skip_conversion:
                    num_cases_skipped += 1
                case.cleanup()

        logging.info(f'Skipped conversion of {num_cases_skipped}')
        end_time = datetime.now()
        logging.info(f'{title} conversion ended at {end_time.isoformat()}\n\t(runtime {end_time - start_time})')



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import json

from picai_prep.data_utils import PathLike
from picai_prep.examples.dcm2mha.sample_archive import \
    generate_dcm2mha_settings
from picai_prep.examples.mha2nnunet.sample_archive import \
    generate_mha2nnunet_settings
from picai_prep.examples.mha2nnunet.sample_archive_inference import \
    generate_mha2nnunet_settings as generate_mha2nnunet_inference_settings


def test_archive_json(
    archive_dir: PathLike = "tests/input/dcm/ProstateX/",
    output_path: PathLike = "tests/output/dcm2mha_settings.json"
):
    """
    Create archive.json for ProstateX sample
    """
    generate_dcm2mha_settings(
        archive_dir=archive_dir,
        output_path=output_path,
    )

    with open(output_path) as fp:
        archive = json.load(fp)

    # compare with expected output
    with open("tests/output-expected/dcm2mha_settings.json") as fp:
        expected_archive = json.load(fp)

    assert archive == expected_archive, "Archive json specification is not as expected!"


def test_mha2nnunet_settings_generation(
    archive_dir: PathLike = "tests/output-expected/mha/ProstateX/",
    output_path: PathLike = "tests/output/mha2nnunet_settings.json"
):
    """
    Create mha2nnunet_settings.json for ProstateX sample
    """
    generate_mha2nnunet_settings(
        archive_dir=archive_dir,
        output_path=output_path
    )

    with open(output_path) as fp:
        archive = json.load(fp)

    # compare with expected output
    with open("tests/output-expected/mha2nnunet_settings.json") as fp:
        expected_archive = json.load(fp)

    assert archive == expected_archive, "Archive json specification is not as expected!"


def test_mha2nnunet_settings_generation_for_inference(
    archive_dir: PathLike = "tests/output-expected/mha/ProstateX/",
    output_path: PathLike = "tests/output/mha2nnunet_inference_settings.json"
):
    """
    Create mha2nnunet_settings.json for ProstateX sample
    """
    generate_mha2nnunet_inference_settings(
        archive_dir=archive_dir,
        output_path=output_path
    )

    with open(output_path) as fp:
        archive = json.load(fp)

    # compare with expected output
    with open("tests/output-expected/mha2nnunet_inference_settings.json") as fp:
        expected_archive = json.load(fp)

    assert archive == expected_archive, "Archive json specification is not as expected!"



mha2nnunet_schema = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "properties": {
        "archive": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "patient_id": {
                        "description": "Custom patient ID. If not provided, uses the 'PatientID' metadata",
                        "type": "string"
                    },
                    "study_id": {
                        "description": "Custom study ID. If not provided, uses the 'StudyInstanceUID' metadata",
                        "type": "string"
                    },
                    "scan_paths": {
                        "description": "Paths to mha collection. Can be relative from --input or absolute",
                        "type": "array",
                        "items": {
                            "type": "string",
                            "uniqueItems": True
                        }
                    },
                    "annotation_path": {
                        "description": "Path to annotation. Can be relative from --input or absolute",
                        "type": "string"
                    },
                    "segmentation_path": {
                        "description": "Path to segmentation. Can be relative from --input or absolute",
                        "type": "string"
                    }
                },
                "required": [
                    "scan_paths"
                ],
                "additionalProperties": False
            }
        },
        "preprocessing": {
            "type": "object",
            "description": "Preprocessing parameters",
            "properties": {
                # preprocess to fixed size (specify at least two),
                # or preprocess to variable size (specify only one)
                "physical_size": {
                    "description": "Target field-of-view in mm (z, y, x). Automatically calculated if `matrix_size` and `spacing` are set.",
                    "$ref": "#/$defs/3d"
                },
                "matrix_size": {
                    "description": "Target matrix size in voxels (z, y, x). Automatically calculated if `physical_size` and `spacing` are set.",
                    "$ref": "#/$defs/3d"
                },
                "spacing": {
                    "description": "Target resolution in mm/voxel (z, y, x). Automatically calculated if `physical_size` and `matrix_size` are set.",
                    "$ref": "#/$defs/3d"
                },
                "crop_only": {
                    "description": "Only crop to specified size (i.e., do not pad)",
                    "type": "boolean"
                },
                "scan_interpolator": {
                    "description": "Interpolator for scans",
                    "type": "integer",
                },
                "lbl_interpolator": {
                    "description": "Interpolator for labels",
                    "type": "integer",
                },
            },
            "additionalProperties": False
        },
        "dataset_json": {
            "type": "object",
            "description": "nnUnet requires a dataset.json file",
            "properties": {
                "task": {
                    "type": "string"
                },
                "name": {
                    "type": "string"
                },
                "description": {
                    "type": "string"
                },
                "tensorImageSize": {
                    "type": "string",
                    "description": "3D or 4D, defaults to 4D",
                    "pattern": "^\\d[Dd]$"
                },
                "reference": {
                    "type": "string",
                    "description": "website of the dataset, if available"
                },
                "licence": {
                    "type": "string"
                },
                "release": {
                    "type": "string"
                },
                "modality": {
                    "$ref": "#/$defs/1a2a3a",
                    "description": "modality names. must be in the same order as the images"
                },
                "labels": {
                    "$ref": "#/$defs/1a2a3a",
                    "description": "..."
                }
            },
            "additionalProperties": True,
            "required": [
                "modality",
                "labels"
            ]
        },
        "options": {
            "type": "object",
            "description": "Additional options for preprocessing",
            "additionalProperties": True
        }
    },
    "additionalProperties": False,
    "required": [
        "archive",
        "preprocessing",
        "dataset_json"
    ],
    "$defs": {
        "3d": {
            "type": "array",
            "minItems": 3,
            "maxItems": 3,
            "items": {
                "type": "number"
            }
        },
        "1a2a3a": {
            "type": "object",
            "patternProperties": {
                "^\\d+$": {
                    "type": "string"
                }
            }
        }
    }
}



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import argparse
import json
import sys

from picai_prep import Dicom2MHAConverter, MHA2nnUNetConverter, nnunet2nndet
from picai_prep.examples import dcm2mha, mha2nnunet


def set_verbosity(args):
    """Set verbosity level in settings"""
    if args.verbose is not None:
        # set verbosity level
        with open(args.json) as fp:
            settings = json.load(fp)
            if "options" not in settings:
                settings["options"] = {}
            settings["options"]["verbose"] = args.verbose
            args.json = settings


def run_dcm2mha(args):
    """Wrapper for DICOM → MHA conversion"""
    set_verbosity(args)

    archive = Dicom2MHAConverter(
        input_dir=args.input,
        output_dir=args.output,
        dcm2mha_settings=args.json,
    )
    archive.convert()


def generate_dcm2mha_settings(args):
    """Wrapper to generate dcm2mha settings"""
    func = {
        "sample_archive": dcm2mha.sample_archive.generate_dcm2mha_settings,
    }[args.structure]

    kwargs = {
        "archive_dir": args.input,
        "output_path": args.json,
    }

    # generate settings
    func(**kwargs)


def run_mha2nnunet(args):
    """Wrapper for MHA → nnUNet conversion"""
    annotations_dir = args.annotations if args.annotations else args.input

    set_verbosity(args)

    archive = MHA2nnUNetConverter(
        output_dir=args.output,
        scans_dir=args.input,
        mha2nnunet_settings=args.json,
        scans_out_dirname=args.scans_out_dirname,
        annotations_dir=annotations_dir,
        annotations_out_dirname=args.annotations_out_dirname,
    )
    archive.convert()

    if not args.no_dataset_json:
        archive.create_dataset_json()


def generate_mha2nnunet_settings(args):
    """Wrapper to generate mha2nnunet settings"""
    func = {
        "picai_archive": mha2nnunet.picai_archive.generate_mha2nnunet_settings,
        "picai_archive_inference": mha2nnunet.picai_archive_inference.generate_mha2nnunet_settings,
        "sample_archive": mha2nnunet.sample_archive.generate_mha2nnunet_settings,
        "sample_archive_inference": mha2nnunet.sample_archive_inference.generate_mha2nnunet_settings
    }[args.structure]

    kwargs = {
        "archive_dir": args.input,
        "output_path": args.json,
    }

    if args.annotations is not None:
        kwargs["annotations_dir"] = args.annotations

    # generate settings
    func(**kwargs)


def run_nnunet2nndet(args):
    """Wrapper for nnUNet → nnDetection conversion"""
    nnunet2nndet(
        nnunet_raw_data_path=args.input,
        nndet_raw_data_path=args.output,
    )


# Set up command line arguments
parser = argparse.ArgumentParser()
subparsers = parser.add_subparsers()

# Argument parser for DICOM → MHA
dcm2mha_parser = subparsers.add_parser('dcm2mha')
dcm2mha_parser.add_argument("-j", "--json", type=str, required=True,
                            help="Path to JSON mappings file")
dcm2mha_parser.add_argument("-i", "--input", type=str, required=True,
                            help="Root directory for input, e.g. /path/to/archive/")
dcm2mha_parser.add_argument("-o", "--output", type=str, required=True,
                            help="Root directory for output")
dcm2mha_parser.add_argument("-v", "--verbose", type=int, required=False,
                            help="Set verbosity: 0 (no logs), 1 (default), 2 (extended logging)")
dcm2mha_parser.set_defaults(func=run_dcm2mha)


# Argument parser for dcm2mha_settings
dcm2mha_settings_parser = subparsers.add_parser('dcm2mha_settings')
dcm2mha_settings_parser.add_argument("-s", "--structure", type=str, required=True,
                                     help="Structure of DICOM Archive")
dcm2mha_settings_parser.add_argument("-i", "--input", type=str, required=True,
                                     help="Path to DICOM Archive")
dcm2mha_settings_parser.add_argument("-j", "--json", type=str, required=True,
                                     help="Path to store dcm2mha_settings.json")
dcm2mha_settings_parser.set_defaults(func=generate_dcm2mha_settings)


# Argument parser for MHA → nnUNet
dcm2mha_parser = subparsers.add_parser('mha2nnunet')
dcm2mha_parser.add_argument("-j", "--json", type=str, required=True,
                            help="Path to JSON mappings file")
dcm2mha_parser.add_argument("-i", "--input", type=str, required=True,
                            help="Path to PICAI .mha data")
dcm2mha_parser.add_argument("-a", "--annotations", type=str, required=False,
                            help="Path to PICAI annotations (defaults to --input)")
dcm2mha_parser.add_argument("-o", "--output", type=str, required=True,
                            help="Root directory for output")
dcm2mha_parser.add_argument("--scans_out_dirname", type=str, default="imagesTr",
                            help="Folder for scans (relative to root directory)")
dcm2mha_parser.add_argument("--annotations_out_dirname", type=str, default="labelsTr",
                            help="Folder for annotations (relative to root directory)")
dcm2mha_parser.add_argument("--no_dataset_json", action='store_true',
                            help="Prevent generating the dataset.json for nnU-Net.")
dcm2mha_parser.add_argument("-v", "--verbose", type=int, default=1,
                            help="Set verbosity: 0 (no logs), 1 (default), 2 (extended logging)")
dcm2mha_parser.set_defaults(func=run_mha2nnunet)


# Argument parser for mha2nnunet_settings
mha2nnunet_settings_parser = subparsers.add_parser('mha2nnunet_settings')
mha2nnunet_settings_parser.add_argument("-s", "--structure", type=str, required=True,
                                        help="Structure of MHA Archive")
mha2nnunet_settings_parser.add_argument("-i", "--input", type=str, required=True,
                                        help="Path to MHA Archive")
mha2nnunet_settings_parser.add_argument("-a", "--annotations", type=str, required=False,
                                        help="Path to PICAI annotations (defaults to --input)")
mha2nnunet_settings_parser.add_argument("-j", "--json", type=str, required=True,
                                        help="Path to store mha2nnunet_settings.json")
mha2nnunet_settings_parser.set_defaults(func=generate_mha2nnunet_settings)


# Argument parser for nnUNet → nnDetection
nnunet2nndet_parser = subparsers.add_parser('nnunet2nndet')
nnunet2nndet_parser.add_argument("-i", "--input", type=str, required=True,
                                 help="Path to nnUNet raw data folder")
nnunet2nndet_parser.add_argument("-o", "--output", type=str, required=True,
                                 help="Path to nnDet raw data folder")
nnunet2nndet_parser.set_defaults(func=run_nnunet2nndet)


if __name__ == '__main__':
    args = parser.parse_args()
    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_help()
        sys.exit(1)
